#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
from GangaCore.Utility.util import unique
from GangaLHCb.Lib.RTHandlers.RTHUtils import lhcbdiracAPI_script_template, lhcbdirac_outputfile_jdl
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import get_share_path, master_sandbox_prepare, sandbox_prepare, script_generator
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_settings, API_nullifier
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from GangaDirac.Lib.Utilities.DiracUtilities import execute
from GangaCore.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import isType
from GangaDirac.Lib.Files.DiracFile import DiracFile
logger = getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class LHCbRootDiracRunTimeHandler(IRuntimeHandler):

    """The runtime handler to run ROOT jobs on the Dirac backend"""

    def master_prepare(self, app, appmasterconfig):
        inputsandbox, outputsandbox = master_sandbox_prepare(
            app, appmasterconfig)
        # check file is set OK
        if not app.script.name:
            msg = 'Root.script.name must be set.'
            raise ApplicationConfigurationError(msg)

        sharedir_scriptpath = os.path.join(get_share_path(app),
                                           os.path.basename(app.script.name))

        if not os.path.exists(sharedir_scriptpath):
            msg = 'Script must exist!'
            raise ApplicationConfigurationError(msg)

        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)
        input_data,   parametricinput_data = dirac_inputdata(app)
        logger.debug("input_data: " + str(input_data))
        job = app.getJobObject()
        outputfiles = [this_file for this_file in job.outputfiles if isType(this_file, DiracFile)]

        lhcb_dirac_outputfiles = lhcbdirac_outputfile_jdl(outputfiles)

        # NOTE special case for replicas: replicate string must be empty for no
        # replication
        params = {'DIRAC_IMPORT': 'from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb',
                  'DIRAC_JOB_IMPORT': 'from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob',
                  'DIRAC_OBJECT': 'DiracLHCb()',
                  'JOB_OBJECT': 'LHCbJob()',
                  'NAME': mangle_job_name(app),
                  'INPUTDATA': input_data,
                  'PARAMETRIC_INPUTDATA': parametricinput_data,
                  'OUTPUT_SANDBOX': API_nullifier(outputsandbox),
                  'OUTPUTFILESSCRIPT' : lhcb_dirac_outputfiles,
                  'OUTPUT_PATH': "",  # job.fqid,
                  'SETTINGS': diracAPI_script_settings(app),
                  'DIRAC_OPTS': job.backend.diracOpts,
                  'PLATFORM': getConfig('ROOT')['arch'],
                  'REPLICATE': 'True' if getConfig('DIRAC')['ReplicateOutputData'] else '',
                  # leave the sandbox for altering later as needs
                  # to be done in backend.submit to combine master.
                  # Note only using 2 #s as auto-remove 3
                  'INPUT_SANDBOX': '##INPUT_SANDBOX##'
                  }

        scriptpath = os.path.join(get_share_path(app),
                                  os.path.basename(app.script.name))

        wrapper_path = os.path.join(job.getInputWorkspace(create=True).getPath(),
                                    'script_wrapper.py')
        python_wrapper =\
"""#!/usr/bin/env python
import os, sys
def formatVar(var):
    try:
        float(var)
        return str(var)
    except ValueError as v:
        return '\\\"%s\\\"' % str(var)

script_args = '###SCRIPT_ARGS###'

del sys.argv[sys.argv.index('script_wrapper.py')]
###FIXARGS###
if script_args == []: script_args = ''
os.system('###COMMAND###' % script_args)
###INJECTEDCODE###
"""

        python_wrapper = python_wrapper.replace('###SCRIPT_ARGS###', str('###JOINER###'.join([str(a) for a in app.args])))

        params.update({ 'APP_NAME' : 'Root',
                        'APP_VERSION' : app.version,
                        'APP_SCRIPT' : wrapper_path,
                        'APP_LOG_FILE' : 'Ganga_Root.log' })

        #params.update({'ROOTPY_SCRIPT': wrapper_path,
        #               'ROOTPY_VERSION': app.version,
        #               'ROOTPY_LOG_FILE': 'Ganga_Root.log',
        #               'ROOTPY_ARGS': [str(a) for a in app.args]})

        f = open(wrapper_path, 'w')
        if app.usepython:
            python_wrapper = script_generator(python_wrapper,
                                              remove_unreplaced=False,
                                              FIXARGS='',
                                              COMMAND='/usr/bin/env python %s %s' % (os.path.basename(app.script.name), '%s'),
                                              JOINER=' ',
                                              #INJECTEDCODE = getWNCodeForOutputPostprocessing(job,'')
                                              )


        else:
            python_wrapper = script_generator(python_wrapper,
                                              remove_unreplaced=False,
                                              FIXARGS='script_args=[formatVar(v) for v in script_args]',
                                              COMMAND='export DISPLAY=\"localhoast:0.0\" && root -l -q \"%s(%s)\"' % (os.path.basename(app.script.name), '%s'),
                                              JOINER=',',
                                              #INJECTEDCODE = getWNCodeForOutputPostprocessing(job,'')
                                              )

        f.write(python_wrapper)
        f.close()

        dirac_script = script_generator(lhcbdiracAPI_script_template(), **params)
        return StandardJobConfig(dirac_script,
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('Root', 'Dirac', LHCbRootDiracRunTimeHandler)





##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: JobRegistry.py,v 1.1.2.1 2009-07-24 13:39:39 ebke Exp $
##########################################################################

#from GangaCore.Utility.external.ordereddict import oDict
from GangaCore.Utility.external.OrderedDict import OrderedDict as oDict

from GangaCore.Core.exceptions import GangaException
from GangaCore.Core.GangaRepository.Registry import Registry, RegistryKeyError, RegistryAccessError, RegistryFlusher

from GangaCore.GPIDev.Base.Proxy import stripProxy, isType

import GangaCore.Utility.logging

from GangaCore.GPIDev.Lib.Job.Job import Job

from .RegistrySlice import RegistrySlice

from .RegistrySliceProxy import RegistrySliceProxy, _wrap, _unwrap

# display default values for job list
from .RegistrySlice import config

logger = GangaCore.Utility.logging.getLogger()


class JobRegistry(Registry):

    def __init__(self, name, doc):
        super(JobRegistry, self).__init__(name, doc)
        self.stored_slice = JobRegistrySlice(self.name)
        self.stored_slice.objects = self
        self.stored_proxy = JobRegistrySliceProxy(self.stored_slice)

    def getSlice(self):
        return self.stored_slice

    def getProxy(self):
        return self.stored_proxy

    def getIndexCache(self, obj):

        cached_values = ['status', 'id', 'name']
        cache = {}
        for cv in cached_values:
            #print("cv: %s" % str(cv))
            cache[cv] = getattr(obj, cv)
            #logger.info("Setting: %s = %s" % (str(cv), str(cache[cv])))
        this_slice = JobRegistrySlice("jobs")
        for dpv in this_slice._display_columns:
            #logger.debug("Storing: %s" % str(dpv))
            try:
                value = this_slice._get_display_value(obj, dpv)
                cache["display:" + dpv] = value
            except Exception as err:
                value = None
        del this_slice

        # store subjob status
        if hasattr(obj, "subjobs"):
            cache["subjobs:status"] = []
            if hasattr(obj.subjobs, "getAllCachedData"):
                for sj in obj.subjobs.getAllCachedData():
                    cache["subjobs:status"].append(sj['status'])
            else:
                for sj in obj.subjobs:
                    cache["subjobs:status"].append(sj.status)

        #print("Cache: %s" % str(cache))
        return cache

    def startup(self):
        """
            This is the main startup method of the Registry
        """
        self._needs_metadata = True
        super(JobRegistry, self).startup()
        if len(self.metadata.ids()) == 0:
            from GangaCore.GPIDev.Lib.JobTree import JobTree
            jt = JobTree()
            stripProxy(jt)._setRegistry(self.metadata)
            self.metadata._add(jt)
        self.jobtree = self.metadata[self.metadata.ids()[-1]]
        self.flush_thread = RegistryFlusher(self, 'JobRegistryFlusher')
        self.flush_thread.start()

    def check(self):
        """
            This code checks for jobs which are in the submitting state
            If a job is 'submitting' an call to force_status('failed') is made
            This should be enough to fix some future issues with monitoring """

        def _shouldAutoKill(this_job):
            """ This checks to see if the job is in a transistional state
                By construction jobs in this state on disk need to be failed
            """
            return this_job.status in ["submitting", "completing"]

        def _shouldAutoCheck(this_job):
            """ This checks to see if the job is in a state where the
                dict describing the job is potentially stale
            """
            return this_job.status in ["new"]

        def _killJob(this_job):
            logger.warning("Auto-Failing job in bad state: %s was %s" % (this_job.getFQID('.'), this_job.status))
            logger.warning("To try this job again resubmit or use a backend.reset()")
            this_job.force_status('failed')

        reverse_job_list = list(self._objects.keys())
        reverse_job_list.reverse()

        num_loaded = 0
        num_checked = 0
        for k in reverse_job_list:
            if (num_loaded == 5) and (num_checked == 5):
                break
            v = self._objects[k]
            if _shouldAutoKill(v):
                if num_loaded == 5:
                    continue
                try:
                    num_loaded+=1
                    self._load(v)
                    logger.debug("Loaded job: %s" % v.getFQID('.'))
                except RegistryLockError:
                    logger.debug("Failed to load job, it's potentially being used by another ganga sesion")
                    continue
                if v.subjobs:
                    haveKilled = False
                    for sj in v.subjobs:
                        if _shouldAutoKill(sj):
                            _killJob(sj)
                            haveKilled = True
                    if not haveKilled:
                        if v.status == 'submitting':
                            v.updateStatus("submitted")
                        logger.warning("Job status re-updated after potential force-close")
                        v.updateMasterJobStatus()
                else:
                    _killJob(v)
            elif _shouldAutoCheck(v):
                if num_checked == 5:
                    continue
                try:
                    num_checked+=1
                    self._load(v)
                    logger.debug("Loaded job: %s" % v.getFQID('.'))
                except RegistryLockError:
                    logger.debug("Failed to load job, it's potentially being used by another ganga sesion")
                    continue


    def shutdown(self):
        self.flush_thread.join()
        super(JobRegistry, self).shutdown()

    def getJobTree(self):
        return self.jobtree

    def _remove(self, obj, auto_removed=0):
        super(JobRegistry, self)._remove(obj, auto_removed)
        try:
            self.jobtree.cleanlinks()
        except Exception as err:
            logger.debug("Exception in _remove: %s" % str(err))
            pass

class JobRegistrySlice(RegistrySlice):

    def __init__(self, name):
        super(JobRegistrySlice, self).__init__(name, display_prefix="jobs")
        from GangaCore.Utility.ColourText import Foreground, Background, Effects
        fg = Foreground()
        fx = Effects()
        bg = Background()
        try:
            status_colours = config['jobs_status_colours']
            self.status_colours = dict([(k, eval(v, {'fx': fx, 'fg': fg, 'bg': bg})) for k, v in status_colours.items()])
        except Exception as x:
            logger.warning('configuration problem with colour specification: "%s"', str(x))
            status_colours = config.options['jobs_status_colours'].default_value
            self.status_colours = dict([(k, eval(v, {'fx': fx, 'fg': fg, 'bg': bg})) for k, v in status_colours.items()])
        self.fx = fx
        self._proxyClass = JobRegistrySliceProxy

    def _getColour(self, obj):
        if isType(obj, Job):
            from GangaCore.GPIDev.Lib.Job.Job import lazyLoadJobStatus
            status_attr = lazyLoadJobStatus(stripProxy(obj))
        elif isType(obj, str):
            status_attr = obj
        elif isType(obj, dict):
            if 'display:status' in obj:
                status_attr = obj['display:status']
            elif 'status' in obj:
                status_attr = obj['status']
            else:
                status_attr = None
        else:
            status_attr = obj
        try:
            returnable = self.status_colours.get(status_attr, self.fx.normal)
        except Exception:
            returnable = self.status_colours.get(self.fx.normal)
        return returnable

    def __call__(self, id):
        """ Retrieve a job by id.
        """
        this_id = id
        t = type(this_id)
        if t is int:
            try:
                return _wrap(self.objects[this_id])
            except KeyError:
                if self.name == 'templates':
                    raise RegistryKeyError('Template %d not found' % this_id)
                else:
                    raise RegistryKeyError('Job %d not found' % this_id)
        elif t is tuple:
            ids = this_id
        elif t is str:
            if this_id.isdigit():
                try:
                    return _wrap(self.objects[int(this_id)])
                except KeyError:
                    if self.name == 'templates':
                        raise RegistryKeyError('Template %d not found' % this_id)
                    else:
                        raise RegistryKeyError('Job %d not found' % this_id)
            elif this_id.count('.') == 1 and id.split('.')[0].isdigit() and this_id.split('.')[1].isdigit():
                ids = this_id.split(".")
            else:
                import fnmatch
                jlist = [j for j in self.objects if fnmatch.fnmatch(j.name, this_id)]
                if len(jlist) == 1:
                    return _wrap(jlist[0])
                return jobSlice(jlist)
        else:
            raise RegistryAccessError('Expected a job id: int, (int,int), or "int.int"')

        if not len(ids) in [1, 2]:
            raise RegistryAccessError('Too many ids in the access tuple, 2-tuple (job,subjob) only supported')

        try:
            ids = [int(this_id) for this_id in ids]
        except TypeError:
            raise RegistryAccessError('Expeted a job id: int, (int,int), or "int.int"')
        except ValueError:
            raise RegistryAccessError('Expected a job id: int, (int,int), or "int.int"')

        try:
            j = self.objects[ids[0]]
        except KeyError:
            if self.name == 'templates':
                raise RegistryKeyError('Template %d not found' % ids[0])
            else:
                raise RegistryKeyError('Job %d not found' % ids[0])

        if len(ids) > 1:
            try:
                return _wrap(j.subjobs[ids[1]])
            except IndexError:
                raise RegistryKeyError('Subjob %s not found' % ('.'.join([str(_id) for _id in ids])))
        else:
            return _wrap(j)

    def submit(self, keep_going):
        self.do_collective_operation(keep_going, 'submit')

    def kill(self, keep_going):
        self.do_collective_operation(keep_going, 'kill')

    def resubmit(self, keep_going):
        self.do_collective_operation(keep_going, 'resubmit')

    def force_status(self, status, keep_going, force):
        self.do_collective_operation(
            keep_going, 'force_status', status, force=force)

    def remove(self, keep_going, force):
        self.do_collective_operation(keep_going, 'remove', force=force)


class JobRegistrySliceProxy(RegistrySliceProxy):

    """This object is an access list of jobs defined in GangaCore. 

    'jobs' represents all existing jobs, so a subset of jobs may 
    be created by slicing: 
    jobs[-10:]      (to get the last ten jobs)

    or selecting:
    jobs.select(status='new')      (to select all new jobs)
    jobs.select(10,20)      (to select those with IDs between 10 and 20)

    A new access list is created as a result of a slice/select operation. 
    The new access list may be further restricted.

    This object allows one to perform collective operations such as
    killing or submiting all jobs in the current range. Setting the optional
    parameter keep_going=True (the default) results in the operation continuing 
    to process all jobs, irrespective of any errors encountered. If keep_going=False 
    then the operation will stop with an Exception at the first error encountered.


    """

    def submit(self, keep_going=True):
        """ Submit all jobs."""
        stripProxy(self).submit(keep_going=keep_going)

    def resubmit(self, keep_going=True):
        """ Resubmit all jobs."""
        return stripProxy(self).resubmit(keep_going=keep_going)

    def kill(self, keep_going=True):
        """ Kill all jobs."""
        return stripProxy(self).kill(keep_going=keep_going)

    def remove(self, keep_going=True, force=False):
        """ Remove all jobs."""
        return stripProxy(self).remove(keep_going=keep_going, force=force)

    def fail(self, keep_going=True, force=False):
        """ Fail all jobs."""
        return stripProxy(self).fail(keep_going=keep_going, force=force)

    def force_status(self, status, keep_going=True, force=False):
        """ Force status of all jobs to 'completed' or 'failed'."""
        return stripProxy(self).force_status(status, keep_going=keep_going, force=force)

    def copy(self, keep_going=True):
        """ Copy all jobs. """
        return JobRegistrySliceProxy(stripProxy(self).copy(keep_going=keep_going))

    def __call__(self, x):
        """ Access individual job. Examples:
        jobs(10) : get job with id 10 or raise exception if it does not exist.
        jobs((10,2)) : get subjobs number 2 of job 10 if exist or raise exception.
        jobs('10.2')) : same as above
        """
        return stripProxy(self).__call__(x)

    def __getitem__(self, x):
        """ Get a job by positional index. Examples:
        jobs[-1] : get last job,
        jobs[0] : get first job,
        jobs[1] : get second job.
        """
        return _wrap(stripProxy(self).__getitem__(_unwrap(x)))

def jobSlice(joblist):
    """create a 'JobSlice' from a list of jobs
    example: jobSlice([j for j in jobs if j.name.startswith("T1:")])"""
    this_slice = JobRegistrySlice("manual slice")
    this_slice.objects = oDict([(j.fqid, _unwrap(j)) for j in joblist])
    return _wrap(this_slice)

# , "Create a job slice from a job list") exported to the Runtime bootstrap




from GangaCore.Runtime.GPIexport import exportToGPI

from GangaCore.Utility.logging import getLogger
logger = getLogger(modulename=True)


def browseBK(gui=True):
    """Return an LHCbDataset from the GUI LHCb Bookkeeping.

    Utility function to launch the new LHCb bookkeeping from inside GangaCore.
    The function returns an LHCbDataset object. 

    After browsing and selecting the desired datafiles, click on the
    \"Save as ...\" button. The Browser will quit and save the seleted files
    as an LHCbDataset object

    Usage:
    # retrieve an LHCbDataset object with the selected files and store
    # them in the variable l
    l = browseBK()

    # retrieve an LHCbDataset object with the selected files and store
    # them in the jobs inputdata field, ready for submission
    j.inputdata=browseBK()    
    """
    import GangaCore.Utility.logging
    from GangaCore.GPIDev.Base.Proxy import addProxy
    logger = GangaCore.Utility.logging.getLogger()
    try:
        from GangaLHCb.Lib.Backends.Bookkeeping import Bookkeeping
        from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
    except ImportError:
        logger.warning('Could not start Bookkeeping Browser')
        return None
    bkk = Bookkeeping()
    return addProxy(bkk.browse(gui))

exportToGPI('browseBK', browseBK, 'Functions')


def fixBKQueryInBox(newCategory='query'):
    import os
    from GangaCore.Utility.Config import getConfig

    def _filt(line):
        return 'class name=\"BKQuery\"' in line

    gangadir = getConfig('Configuration')['gangadir']
    logger.info('found gangadir = ' + gangadir)
    for root, dirs, files in os.walk(gangadir):
        if 'data' in files and 'box' in root and not 'box.' in root:
            path = os.path.join(root, 'data')
            logger.info("looking at " + path)
            f1 = open(path, 'r')
            f2 = open(path + '~', 'r')
            lines1 = f1.readlines()
            lines2 = f2.readlines()
            line1 = list(filter(_filt, lines1))
            line2 = list(filter(_filt, lines2))
            f1.close()
            f2.close()

            newline = ' <class name=\"BKQuery\" version=\"1.2\" category=\"%s\">\n' % newCategory
            if len(line1) is 1:
                lines1[lines1.index(line1[0])] = newline
                logger.info('backing up old settings...')
                os.system('cp %s %sX; mv %s~ %s~X' % (path, path, path, path))
                f1 = open(path, 'w')
                f2 = open(path + '~', 'w')
                for l in lines1:
                    f1.write(l)
                    f2.write(l)
                f1.close()
                f2.close()
                p = os.path.join(root[:-2], root.split('/')[-1] + '.index')
                os.system('rm -f %s' % p)
            elif len(line2) is 1:
                lines2[lines2.index(line2[0])] = newline
                logger.info('backing up old settings...')
                os.system('cp %s %sX; mv %s~ %s~X' % (path, path, path, path))
                f1 = open(path, 'w')
                f2 = open(path + '~', 'w')
                for l in lines2:
                    f1.write(l)
                    f2.write(l)
                f1.close()
                f2.close()
                p = os.path.join(root[:-2], root.split('/')[-1] + '.index')
                os.system('rm -f %s' % p)
    logger.info("box repository converted!\n")
    logger.info(
        "PLEASE NOW QUIT THIS GANGA SESSION AND RESTART TO SEE EFFECTS.")

exportToGPI('fixBKQueryInBox', fixBKQueryInBox, 'Functions')


def restoreOLDBox():
    import os
    from GangaCore.Utility.Config import getConfig
    gangadir = getConfig('Configuration')['gangadir']
    logger.info('found gangadir = ' + gangadir)
    for root, dirs, files in os.walk(gangadir):
        if 'dataX' in files and 'box' in root and not 'box.' in root:
            path = os.path.join(root, 'data')
            logger.info("restoring old BKQuery box file...")
            os.system('mv %sX %s; mv %s~X %s~' % (path, path, path, path))

    logger.info("box repository converted!\n")
    logger.info(
        "PLEASE NOW QUIT THIS GANGA SESSION AND RESTART TO SEE EFFECTS.")

exportToGPI('restoreOLDBox', restoreOLDBox, 'Functions')



#!/usr/bin/env python3

from __future__ import print_function

# Ganga tutorial application developed by Kuba Moscicki, CERN, 2006

# Find all prime factors of a given NUMBER using a list of primes stored in a text FILE.
# Prime factors are written to a *.dat file in the following way:
# [(p1,n1), (p2,n2), ..., (pk,nk)]
#
# p are prime factors and n are multipliers i.e. p1^n1 * p2^n2 * ... * pk^nk is a factor of the NUMBER
#
# Arguments: NUMBER FILE
#
# The FILE has the following format:
#                  The First 1,000,000 Primes (from primes.utm.edu)
#
#         2         3         5         7        11        13        17        19
#        23        29        31        37        41        43        47        53
#        59        61        67        71        73        79        83        89
#        97       101       103       107       109       113       127       131
#       137       139       149       151       157       163       167       173
#       179       181       191       193       197       199       211       223
#       227       229       233       239       241       251       257       263
#       269       271       277       281       283       293       307       311
#
# If URL (http address) is given, then file is downloaded and unzipped first.
#

import sys, os
import requests

NUMBER = int(sys.argv[1])

pfns = sys.argv[2:]

factors = []

currDir = os.path.dirname(os.path.abspath(__file__))

def download_url(url, save_path, chunk_size=128):
    r = requests.get(url, stream=True, verify=False)
    with open(save_path, 'wb') as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)

for pfn in pfns:
    # download primes file from the web server, unzip and replace extension from .zip to .txt
    # print(pfn)
    if pfn.find('http://') != -1 or pfn.find('https://') != -1:
        download_url(pfn, os.path.join(currDir, os.path.basename(pfn)))
        pfn = os.path.basename(pfn)
        os.system('unzip %s'%pfn)
        pfn = os.path.splitext(pfn)[0]+'.txt'

    pf = open(pfn)

    # skip two first lines of the file
    pf.readline()
    pf.readline()

    line = pf.readline()

    # scan all lines in the file and try to divide the NUMBER by consecutive primes
    while line:
        primes = [int(x) for x in line.split()]

        for p in primes:
            k = 0
            n = NUMBER
            while n%p == 0:
                k += 1
                n /= p

            if k>0:
                factors.append((p,k))
    
        line = pf.readline()

print('Prime factors:',factors)

# check if all prime factors have been found
import math
check = 1
for f in factors:
    check *= int(math.pow(f[0],f[1]))

if int(check) == NUMBER:
    print('All prime factors found!')
else:
    print('Some prime factors are still to be found. Known factors multiply to',check)

# write the factors to a data file
ofn = 'factors-%d.dat'%NUMBER
of = open(ofn,'w')
of.write(str(factors))
print('Created data file',ofn)



#!/usr/bin/env ganga

from __future__ import print_function

# Ganga tutorial application developed by Kuba Moscicki, CERN, 2006

# Helper function for prime number factorization application.

def check_prime_job(j):
    """ Check the output of job j (or subjobs if exist) and print all calculated prime factors.
    """
    NUMBER = j.application.number

    print('Looking for prime factors of',NUMBER)
    
    # we can use our function for subjobs or simple jobs at the same time
    if j.subjobs:
        rjobs = j.subjobs
    else:
        rjobs = [j]

    # loop over output files and build the factor tuples
    factors = []
    for s in rjobs:
        try:
            fc = eval(open(s.outputdir+'/factors-%d.dat'%NUMBER).read())
            if fc:
                print('job',s.id,': got factors:',fc)
            factors.extend(fc)
        except IOError as x:
            print(x)

    # check if we have all prime numbers!
    import math
    check = 1
    for f in factors:
            check *= int(math.pow(f[0],f[1]))
    print
    
    if check == NUMBER:
        print('All prime factors found:',factors)
    else:
        print('Factors found so far',factors,'mulitply to',check)
        print('There are still missing factors!')


def split_prime_job(j,millions,number=None):
    """ Create a splitter for job j to search a given number of millions of prime numbers.
    A number to factorize may be specified else the one defined before is taken. This is useful if you make a copy of a job.
    Example: split_prime_job(j,5,123456) - search first 5 millions of prime numbers and find factors of 123456
    """

    if not number:
        number = j.application.args[0]

    from GangaCore.GPI import File, TUTDIR, ArgSplitter, Executable
    j.application.exe=File(TUTDIR+'/Lib/primes/prime_factor.py')

    # make sure that we crab the data files!
    j.outputsandbox = ['*.dat']

    j.application.args = [str(number)]
    j.splitter = ArgSplitter()
    args = []
    for m in range(1,millions+1):
        args.append([str(number),'https://primes.utm.edu/lists/small/millions/primes%d.zip'%m])
    j.splitter.args = args




# PhysicalFile is now a pseudonym for the LocalFile... This is for backwards
# compatability in the simplest of cases, where the user needs to do something more
# complicated they're encouraged to update their code

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


import os
import re
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from GangaCore.Utility.files import expandfilename
from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Base.Proxy import GPIProxyObjectFactory
from GangaCore.GPIDev.Lib.File.LocalFile import LocalFile
import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def full_expand_filename(name):
    if len(name) >= 4 and name[0:4].upper() == 'LFN:':
        msg = 'Can not create PhysicalFile from string that begins w/ "LFN:".'\
              ' You probably want to create a DiracFile.'
        raise GangaException(msg)
    urlprefix = re.compile(r'^(([a-zA-Z_][\w]*:)+/?)?/')
    if len(name) >= 4 and name[0:4].upper() == 'PFN:':
        name = name[4:]
    expanded_name = expandfilename(name)
    if urlprefix.match(expanded_name):
        return expanded_name
    return os.path.abspath(expanded_name)

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class PhysicalFile(LocalFile):

    '''Class for handling physical files (i.e. PFNs)

    Example Usage:
    pfn = PhysicalFile("/some/pfn.file")
    pfn.upload("/some/lfn.file","CERN-USER") # upload the PFN to LFC
    [...etc...]
    '''
    _schema = Schema(Version(1, 0), {'name': SimpleItem(defvalue='', doc='PFN'),
                                     'namePattern': SimpleItem(defvalue="", doc='pattern of the file name', transient=1),
                                     'localDir': SimpleItem(defvalue="", doc='local dir where the file is stored, used from get and put methods', transient=1),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1, typelist=['GangaCore.GPIDev.Lib.File.LocalFile'],
                                                               sequence=1, copyable=0, doc="collected files from the wildcard namePattern", transient=1),
                                     'compressed': SimpleItem(defvalue=False, typelist=['bool'], protected=0, doc='wheather the output file should be compressed before sending somewhere', transient=1)})
    _category = 'gangafiles'
    _name = 'PhysicalFile'
    _exportmethods = ['location', 'upload']

    def __init__(self, name=''):
        val = full_expand_filename(name)
        super(PhysicalFile, self).__init__(namePattern=val)
        self.namePattern = os.path.basename(name)
        self.localDir = os.path.dirname(val)
        self.name = val
        logger.warning(
            "!!! PhysicalFile has been deprecated, this is now just a wrapper to the LocalFile object")
        logger.warning(
            "!!! Please update your scripts before PhysicalFile is removed")

    def _attribute_filter__set__(self, n, v):
        if n == 'name':
            import os.path
            val = full_expand_filename(v)
            self.name = val
            self.namePattern = os.path.basename(val)
            self.localDir = os.path.dirname(val)
            return val
        return v

    def upload(self, lfn, diracSE, guid=None):

        from GangaDirac.Lib.Files.DiracFile import DiracFile
        diracFile = DiracFile(namePattern=self.name, lfn=lfn)

        diracFile.put(force=True)

        return diracFile

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



"""
Sandbox functions used in the job wrapper script on the worker node.
The text of this module is sourced into the job wrapper script.
It therefore may use ###TAGS###  which are expanded in the wrapper script.
"""

INPUT_TARBALL_NAME = '_input_sandbox.tgz'
OUTPUT_TARBALL_NAME = '_output_sandbox.tgz'
PYTHON_DIR = '_python'

import os
import mimetypes

import tarfile
from contextlib import closing

def multi_glob(pats, exclude=None):
    """ glob using a list of patterns and removing duplicate files, exclude name in the list for which the callback exclude(name) return true
    example: advanced_glob(['*.jpg','*.gif'],exclude=lambda n:len(n)>20) return a list of all JPG and GIF files which have names shorter then 20 characters
    """

    unique = {}
    if exclude is None:
        def exclude(n): return 0

    for p in pats:
        for f in glob.glob(p):
            unique[f] = 1

    return [name for name in unique.keys() if not exclude(name)]


def recursive_copy(src, dest):
    """ copy src file (or a directory tree if src specifies a directory) to dest directory. dest must be a directory and must exist.
    if src is a relative path, then the src directory structure is preserved in dest.
    """

    if not os.path.isdir(dest):
        raise ValueError(
            'resursive_copy: destination %s must specify a directory (which exists)' % dest)

    if os.path.isdir(src):
        destdir = dest
        srcdir, srcbase = os.path.split(src.rstrip('/'))
        if not srcdir == '' and not os.path.isabs(src):
            destdir = os.path.join(destdir, srcdir)
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
        shutil.copytree(src, os.path.join(destdir, srcbase))
    else:

        srcdir = os.path.dirname(src.rstrip('/'))
        if srcdir == '' or os.path.isabs(src):
            shutil.copy(src, dest)
        else:
            destdir = os.path.join(dest, srcdir)
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
            shutil.copy(src, destdir)

def getPackedInputSandbox(tarpath, dest_dir='.'):
    """Get all sandbox_files from tarball and write them to the workdir.
       This function is called by wrapper script at the run time.
    Arguments:
      'tarpath': a path to the tarball
      'dest_dir': a destination directory
    """

    try:
        with closing(tarfile.open(tarpath, "r:*")) as tf:
            tf.extractall(dest_dir)
    except:
        raise Exception("Error opening tar file: %s" % tarpath)


def createOutputSandbox(output_patterns, filter, dest_dir):
    """Get all files matching output patterns except filtered with filter and
       write them to the destination directory.
       This function is called by wrapper script at the run time.
    Arguments:
      'output_patterns': list of filenames or patterns.
      'filter': function to filter files (return True to except) 
      'dest_dir': destination directory for output files
    """

    for f in multi_glob(output_patterns, filter):
        try:
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)
            recursive_copy(f, dest_dir)
        except Exception as x:
            print("ERROR: (job ###JOBID### createOutput )", x)


def createPackedOutputSandbox(output_patterns, _filter, dest_dir):
    """Get all files matching output patterns except filtered with filter and
       put them to the Sandbox tarball in destination directory.
       This function is called by wrapper script at the run time.
    Arguments:
      'output_patterns': list of filenames or patterns.
      'filter': function to filter files (return True to except) 
      'dest_dir': destination directory for tarball
    """

    tgzfile = os.path.join(dest_dir, OUTPUT_TARBALL_NAME)

    outputlist = multi_glob(output_patterns, _filter)

    if outputlist:
        if mimetypes.guess_type(tgzfile)[1] in ['gzip']:
            file_format = 'gz'
        elif mimetypes.guess_type(tgzfile)[1] in ['bzip2']:
            file_format = 'bz2'
        else:
            file_format = ''
        with closing(tarfile.open(tgzfile, "w:%s" % file_format)) as tf:
            tf.dereference = True
            for f in outputlist:
                tf.add(f)





def standardSetup():
    import GangaTest.PACKAGE
    PACKAGE.standardSetup()


def loadPlugins( config = None ):
    import GangaTest.Framework.runner
    import GangaTest.Lib.GListApp
    import GangaTest.Lib.TFile
    import GangaTest.Lib.TestApplication
    import GangaTest.Lib.TestSubmitter # TestSubmitter, TestSplitter here
    import GangaTest.Lib.TestRobot
    import GangaTest.Lib.TestObjects
    import GangaTest.Framework.runner





################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id$
################################################################################
                                                                                                              

import os, sys, time, commands, re, tempfile, exceptions, urllib, datetime
import cPickle as pickle

from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Adapters.IBackend import IBackend
from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Job import JobStatusError
from GangaCore.Core.exceptions import BackendError
from GangaCore.Core import Sandbox
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.Core import FileWorkspace
from GangaCore.Utility.Shell import Shell
from GangaCore.Utility.Config import makeConfig, ConfigError, getConfig, setConfigOption
from GangaCore.Utility.logging import getLogger

from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import ToACache
from GangaAtlas.Lib.ATLASDataset.ATLASDataset import Download

from GangaAtlas.Lib.Credentials.ProxyHelper import getNickname

logger = getLogger()
config = getConfig('Jedi')

def retrieveMergeJobs(job, pandaJobDefId):
    '''
    methods for retrieving panda job ids of merging jobs given a jobDefId
    '''
    from pandatools import Client

    ick       = False
    status    = ''
    num_mjobs = 0

    (ec, info) = Client.checkMergeGenerationStatus(pandaJobDefId)

    if ec == 0:

        try:
            status         = info['status']
            mergeJobDefIds = info['mergeIDs']

            if status == 'NA':
                logger.warning('No merging jobs expected')
                job.backend.mergejobs = []

            elif status == 'generating':
                logger.debug('merging jobs are generating')
                job.backend.mergejobs = []

            elif status == 'standby':
                logger.debug('merging jobs to be created')
                job.backend.mergejobs = []

            elif status == 'generated':
                logger.debug('merging jobs are generated')

                for id in mergeJobDefIds:
                    logger.debug("merging jobDefId: %d" % id)

                    ## retrieve merging job id,status given the jobDefId
                    (ec2, mjs) = Client.getPandIDsWithJobID(id)

                    if ec2 == 0:

                        for jid,jinfo in mjs.items():
                            mjobj = PandaMergeJob()
                            mjobj.id     = jid
                            #mjobj.status = jinfo[0]
                            mjobj.url    = 'http://panda.cern.ch/?job=%d' % jid

                            if mjobj not in job.backend.mergejobs:
                                job.backend.mergejobs.append(mjobj)
                            else:
                                logger.debug("merging job %s already exists locally" % mjobj.id)

                            num_mjobs += 1
                    else:
                        logger.warning("getPandIDsWithJobID returns non-zero exit code: %d" % ec2)

            ick = True

        except KeyError:
            logger.error('unexpected job information: %s' % repr(info))

        except Exception as e:
            logger.error('general merge job information retrieval error')
            raise e

    else:
        logger.error('checkMergeGenerationStatus returns non-zero exit code: %d' % ec)

    return (ick, status, num_mjobs)

def retrievePandaJobs(job, jIDs):
    '''
    methods for retrieving panda job ids of panda jobs given a jobDefId
    '''
    from pandatools import Client

    ick       = False
    jstatus    = ''
    num_pjobs = 0

    logger.debug("retrievePandaJobs jIDs=%s" %jIDs)

    # get status from Panda server
    rc, jobsStatus = Client.getFullJobStatus(jIDs,False)
    if rc:
        logger.error('Return code %d retrieving job status information.',rc)
        raise BackendError('Jedi','Return code %d retrieving job status information.' % rc)

    for status in jobsStatus:
        if not status: continue

        jstatus = status.jobStatus
        if status.jobStatus is None:
            logger.warning('No panda jobs expected')
            job.backend.pandajobs = []

        elif status.jobStatus in [ "defined", "activated", "running", "failed", "finished", "holding", "assigned"]:
            logger.debug('Panda jobs are running')
            logger.debug("PandaID: %d" % status.PandaID)

            pjobj         = JediPandaJob()
            pjobj.id      = status.PandaID
            pjobj.url     = 'http://panda.cern.ch/?job=%d' % status.PandaID
            pjobj.jopSpec = dict(zip(status._attributes,status.values()))
            for k in pjobj.jobSpec.keys():
                if type(pjobj.jobSpec[k]) not in [type(''),type(1)]:
                    pjobj.jobSpec[k]=str(pjobj.jobSpec[k])

            if pjobj not in job.backend.pandajobs:
                job.backend.pandajobs.append(pjobj)
            else:
                logger.debug("Panda job %s already exists locally" % pjobj.id)
                
            num_pjobs += 1
        else:
            logger.warning("getFullJobStatus returned unsupported status %s for Panda job %s " %(status.jobStatus, status.PandaID) )
            
        ick = True

    return (ick, jstatus, num_pjobs)

def checkForRebrokerage(string):
    import re
    matchObj = re.match('reassigned to another site by rebrokerage. new PandaID=(\d+) JobsetID=(\d+) JobID=(\d+)', string)
    if matchObj:
        newPandaID = long(matchObj.group(1))
        newJobsetID = long(matchObj.group(2))
        newJobID = long(matchObj.group(3))
        return newPandaID
    raise BackendError('Jedi','Error getting new PandaID for rebrokered job. Report to DA Help')

class JediPandaMergeJob(GangaObject):
    _schema = Schema(Version(1,0), {
        'id'            : SimpleItem(defvalue=None,typelist=['type(None)','int'],protected=0,copyable=0,doc='Panda Job id'),
        'status'        : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=0,copyable=0,doc='Panda Job status'),
        'jobSpec'       : SimpleItem(defvalue={},optional=1,protected=1,copyable=0,doc='Panda JobSpec'),
        'url'           : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=1,copyable=0,doc='Web URL for monitoring the job.')
    })

    _category = 'JediPandaMergeJob'
    _name = 'JediPandaMergeJob'

    def __init__(self):
        super(JediPandaMergeJob,self).__init__()

    def __eq__(self, other):
        return other.id == self.id

class JediPandaJob(GangaObject):
    _schema = Schema(Version(1,0), {
        'id'            : SimpleItem(defvalue=None,typelist=['type(None)','int'],protected=0,copyable=0,doc='Panda Job id'),
        'status'        : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=0,copyable=0,doc='Panda Job status'),
        'jobSpec'       : SimpleItem(defvalue={},optional=1,protected=1,copyable=0,doc='Panda JobSpec'),
        'url'           : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=1,copyable=0,doc='Web URL for monitoring the job.'),
        'exitcode'      : SimpleItem(defvalue='',protected=1,copyable=0,doc='Application exit code (transExitCode)'),
        'piloterrorcode': SimpleItem(defvalue='',protected=1,copyable=0,doc='Pilot Error Code'),
        'reason'        : SimpleItem(defvalue='',protected=1,copyable=0,doc='Error Code Diagnostics'),
    })

    _category = 'JediPandaJob'
    _name = 'JediPandaJob'
    _exportmethods = ['get_stats']

    def __init__(self):
        super(JediPandaJob,self).__init__()

    def __eq__(self, other):
        return other.id == self.id

    def get_stats(self):
        fields = {
            'site':"self.jobSpec['computingSite']",
            'exitstatus':"self.jobSpec['transExitCode']",
            'outse':"self.jobSpec['destinationSE']",
            'jdltime':"''",
            'submittime':"int(time.mktime(time.strptime(self.jobSpec['creationTime'],'%Y-%m-%d %H:%M:%S')))",
            'starttime':"int(time.mktime(time.strptime(self.jobSpec['startTime'],'%Y-%m-%d %H:%M:%S')))",
            'stoptime':"int(time.mktime(time.strptime(self.jobSpec['endTime'],'%Y-%m-%d %H:%M:%S')))",
            'totalevents':"int(self.jobSpec['nEvents'])", 
            'nInputDataFiles':"int(self.jobSpec['nInputDataFiles'])", 
            'inputFileType':"int(self.jobSpec['inputFileType'])", 
            'inputFileProject':"self.jobSpec['inputFileProject']", 
            'inputFileBytes':"int(self.jobSpec['inputFileBytes'])", 
            'nOutputDataFiles':"int(self.jobSpec['nOutputDataFiles'])", 
            'outputFileBytes':"int(self.jobSpec['outputFileBytes'])", 
            'jobMetrics':"self.jobSpec['jobMetrics']", 
            'wallclock':"(int(time.mktime(time.strptime(self.jobSpec['endTime'],'%Y-%m-%d %H:%M:%S')))-int(time.mktime(time.strptime(self.jobSpec['startTime'],'%Y-%m-%d %H:%M:%S'))))",
            'percentcpu':"int(100*self.jobSpec['cpuConsumptionTime']/float(self.jobSpec['cpuConversion'])/(int(time.mktime(time.strptime(self.jobSpec['endTime'],'%Y-%m-%d %H:%M:%S')))-int(time.mktime(time.strptime(self.jobSpec['startTime'],'%Y-%m-%d %H:%M:%S')))))",
            'numfiles':'""',
            'gangatime1':'""',
            'gangatime2':'""',
            'gangatime3':'""',
            'gangatime4':'""',
            'gangatime5':'""',
            'pandatime1':"int(self.jobSpec['pilotTiming'].split('|')[0])",
            'pandatime2':"int(self.jobSpec['pilotTiming'].split('|')[1])",
            'pandatime3':"int(self.jobSpec['pilotTiming'].split('|')[2])",
            'pandatime4':"int(self.jobSpec['pilotTiming'].split('|')[3])",
            'pandatime5':"int(self.jobSpec['pilotTiming'].split('|')[4])",
            'NET_ETH_RX_PREATHENA':'""',
            'NET_ETH_RX_AFTERATHENA':'""'
            }
        stats = {}
        for k in fields.keys():
            try:
                stats[k] = eval(fields[k])
            except:
                pass
        return stats

class Jedi(IBackend):
    '''Jedi backend: submission to the Jedi/PanDA workload management system
    '''

    _schema = Schema(Version(1,0), {
        'site'          : SimpleItem(defvalue='AUTO',protected=0,copyable=1,doc='Require the job to run at a specific site'),
        'requirements'  : ComponentItem('JediRequirements',doc='Requirements for the resource selection'),
        'extOutFile'    : SimpleItem(defvalue=[],typelist=['str'],sequence=1,protected=0,copyable=1,doc='define extra output files, e.g. [\'output1.txt\',\'output2.dat\']'),        
        'id'            : SimpleItem(defvalue=None,typelist=['type(None)','int'],protected=1,copyable=0,doc='PandaID of the job'),
        'url'           : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=1,copyable=0,doc='Web URL for monitoring the job.'),
        'status'        : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=1,copyable=0,doc='Panda job status'),
        'actualCE'      : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=1,copyable=0,doc='Actual CE where the job is run'),
        'libds'         : SimpleItem(defvalue=None,typelist=['type(None)','str'],protected=0,copyable=1,doc='Existing Library dataset to use (disables buildjob)'),
        'buildjob'      : ComponentItem('PandaBuildJob',load_default=0,optional=1,protected=1,copyable=0,doc='Panda Build Job'),
        'buildjobs'     : ComponentItem('PandaBuildJob',sequence=1,defvalue=[],optional=1,protected=1,copyable=0,doc='Panda Build Job'),
        'mergejobs'     : ComponentItem('JediPandaMergeJob',sequence=1,defvalue=[],optional=1,protected=1,copyable=0,doc='Panda Output Merging Jobs'),
        'pandajobs'     : ComponentItem('JediPandaJob',sequence=1,defvalue=[],optional=1,protected=1,copyable=0,doc='Jedi Panda Jobs'),
        'jobSpec'       : SimpleItem(defvalue={},optional=1,protected=1,copyable=0,doc='Panda JobSpec'),
        'exitcode'      : SimpleItem(defvalue='',protected=1,copyable=0,doc='Application exit code (transExitCode)'),
        'piloterrorcode': SimpleItem(defvalue='',protected=1,copyable=0,doc='Pilot Error Code'),
        'reason'        : SimpleItem(defvalue='',protected=1,copyable=0,doc='Error Code Diagnostics'),
        'accessmode'    : SimpleItem(defvalue='',protected=0,copyable=1,doc='EXPERT ONLY'),
        'forcestaged'   : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Force staging of input DS'),
        'individualOutDS': SimpleItem(defvalue=False,protected=0,copyable=1,doc='Create individual output dataset for each data-type. By default, all output files are added to one output dataset'),
        'bexec'         : SimpleItem(defvalue='',protected=0,copyable=1,doc='String for Executable make command - if filled triggers a build job for the Execuatble'),
        'nobuild'       : SimpleItem(defvalue=False,protected=0,copyable=1,doc='Boolean if no build job should be sent - use it together with Athena.athena_compile variable'),
    })

    _category = 'backends'
    _name = 'Jedi'
    _exportmethods = ['get_stats']
  
    def __init__(self):
        super(Jedi,self).__init__()

    def master_submit(self,rjobs,subjobspecs,buildjobspec):
        '''Submit jobs'''
       
        from pandatools import Client
        from pandatools import MiscUtils

        from GangaCore.Core.exceptions import IncompleteJobSubmissionError
        from GangaCore.Utility.logging import log_user_exception

        job = self.getJobObject()

        # job name
        jobName = 'ganga.%s' % MiscUtils.wrappedUuidGen()

        jobspecs = {}
        if buildjobspec:
            jobspecs = buildjobspec
        else:
            jobspecs = subjobspecs

        logger.debug(jobspecs)

        # submit task
        for subjob in rjobs:
            subjob.updateStatus('submitting')

        logger.info("Submitting to Jedi ...")
        verbose = logger.isEnabledFor(10)
        status, tmpOut = Client.insertTaskParams(jobspecs, verbose)

        logger.debug(tmpOut)

        if status != 0:
            logger.error("Task submission to Jedi failed with %s " %status)
            return False
        if tmpOut[0] == False:
            logger.error("Task submission to Jedi failed %s" %tmpOut[1])
            return False
        logger.info("Task submission to Jedi suceeded with new jediTaskID=%s" %tmpOut[1])

        #if buildjobspec:
        #    job.backend.buildjob = PandaBuildJob() 
        #    job.backend.buildjob.id = jobids[0][0]
        #    job.backend.buildjob.url = 'http://panda.cern.ch/?job=%d'%jobids[0][0]
        #    del jobids[0]

        for subjob in rjobs:
            subjob.backend.id = tmpOut[1]
            subjob.backend.url = 'http://pandamon.cern.ch/jedi/taskinfo?days=20&task=%d'%tmpOut[1]
            subjob.updateStatus('submitted')
            logger.info("Panda monitor url: %s" %subjob.backend.url)

        return True

    def master_kill(self):
        '''Kill jobs'''  

        from pandatools import Client

        job = self.getJobObject()
        logger.debug('Killing job %s' % job.getFQID('.'))

        active_status = [ None, 'registered', 'waiting', 'defined', 'pending', 'assigning', 'ready', 'scouting', 'running', 'holding', 'merging', 'prepared', 'aborting', 'finishing' ]
        #active_status = [ None, 'defined', 'unknown', 'assigned', 'waiting', 'activated', 'sent', 'starting', 'running', 'holding', 'transferring' ]

        if self.id and self.status in active_status: 
            status, output = Client.killTask(self.id)
            if status:
                logger.error('Failed killing job (status = %d)',status)
                return False
            else:
                logger.info('Killing Jedi task %s, Server returned: %s' %(self.id, output))
        else:
            logger.error('Cannot kill Jedi job %s since it is not in active status', self.id)

        return True

    @staticmethod
    def master_updateMonitoringInformation(jobs):
        '''Monitor jobs'''       
        from pandatools import Client

        #active_status = [ None, 'defined', 'unknown', 'assigned', 'waiting', 'activated', 'sent', 'starting', 'running', 'holding', 'transferring' ]

        submitting_status = [ ]
        active_status = [ None, 'registered', 'waiting', 'defined', 'pending', 'assigning', 'ready', 'scouting', 'running', 'holding', 'merging', 'prepared', 'aborting', 'finishing' ]
 
        inactive_status = [ 'finished', 'aborted', 'broken', 'failed', 'done' ]

        # Find jobs to be monitored
        jobdict = {}
        for job in jobs:
            # add a delay as Panda can be a little slow in sorting out a new Task
            if job.backend.id and job.backend.status in active_status and ( (datetime.datetime.utcnow() - job.time.timestamps["submitted"]).seconds > 120):
                jobdict[job.backend.id] = job 

        logger.debug("jobdict = %s" %jobdict)
        
        # Monitor active Jedi tasks
        allJobIDs = jobdict.keys()
        pandaJobIDs = {}
        for jID in allJobIDs:
            status, jediTaskDict = Client.getJediTaskDetails({'jediTaskID': jID},False,True,verbose=False)
            if status != 0:
                logger.error("Failed to get task details for %s" % jID)
                #raise BackendError('Jedi','Return code %d retrieving job status information.' % status)
                continue
            # Retrieve job
            job = jobdict[jediTaskDict['jediTaskID']]
            # Store associated Panda jobs
            if job.backend.pandajobs:
                pandaJobIDs[job.backend.id] = [pj.id for pj in job.backend.pandajobs]
            else:
                pandaJobIDs[jediTaskDict['jediTaskID']] = jediTaskDict['PandaID']
            logger.debug("jID = %s, pandaJobIDs = %s" % (jID, pandaJobIDs))

            # Fill the output data dataset list
            if 'outDS' in jediTaskDict and jediTaskDict['outDS'] != '':
                for ds in jediTaskDict['outDS'].split(','):
                    if not ds in job.outputdata.datasetList:
                        job.outputdata.datasetList.append(ds)

            # Jedi job status has changed
            if job.backend.status != jediTaskDict['status']:
                logger.debug('Job %s has changed status from %s to %s',job.getFQID('.'),job.backend.status, jediTaskDict['status'])
                job.backend.status = jediTaskDict['status']
                job.backend.reason = jediTaskDict['statistics']

                # Now update Jedi job status
                if jediTaskDict['status'] in ['registered', 'waiting', 'defined', 'pending', 'assigning', 'ready']:
                    job.updateStatus('submitted')
                elif jediTaskDict['status'] in ['scouting', 'running', 'holding', 'merging', 'prepared' ]:
                    job.updateStatus('running')
                elif jediTaskDict['status'] in ['done']:
                    job.updateStatus('completed')
                elif jediTaskDict['status'] in ['failed', 'finished']:
                    job.updateStatus('failed')
                elif jediTaskDict['status'] in [ 'aborted', 'broken', 'cancelled' ] and job.status not in ['completed','failed']:
                    job.updateStatus('killed')
                else:
                    logger.warning('Unexpected Jedi task status %s', jediTaskDict['status'])

            # Check if associated Panda job exist and monitor them
            if not job.backend.pandajobs:
                jdefids = pandaJobIDs[jID]
                # skip if there are no Panda jobs yet 
                if not jdefids:
                    continue
                tot_num_mjobs = 0

                do_master_update = True
                ick,status,num_mjobs = retrievePandaJobs(job, jdefids)
                logger.debug('retrievePandaJobs returns: %s %s' % (repr(ick),status))
                if not ick:
                    logger.debug('Panda job retrival failure for Jedi task %s with PandaIds %s' % (job.backend.id, jdefids))
                    do_master_update = False

                tot_num_mjobs += num_mjobs
                logger.debug('Job %s retrieved %d Panda jobs' % (job.getFQID('.'),tot_num_mjobs) )
            # Now monitor the already attached Panda jobs
            else:
                jdefids = [ pj.id for pj in job.backend.pandajobs ] 
                rc, jobsStatus = Client.getFullJobStatus(jdefids,False)
                if rc:
                    logger.error('Return code %d retrieving job status information.',rc)
                    raise BackendError('Jedi','Return code %d retrieving job status information.' % rc)

                for status in jobsStatus:
                    if not status: continue

                    for pjob in job.backend.pandajobs:
                        if pjob.id == status.PandaID:
                            # skip if no status change
                            if pjob.status == status.jobStatus:
                                continue 
                            # Else update job record
                            pjob.jobSpec = dict(zip(status._attributes,status.values()))

                            for k in pjob.jobSpec.keys():
                                if type(pjob.jobSpec[k]) not in [type(''),type(1)]:
                                    pjob.jobSpec[k]=str(pjob.jobSpec[k])

                            logger.debug('Job %s with Panda job %s has changed status from %s to %s',job.getFQID('.'),pjob.id, pjob.status,status.jobStatus)
                            pjob.status = status.jobStatus
                            pjob.exitcode = str(status.transExitCode)
                            pjob.piloterrorcode = str(status.pilotErrorCode)
                            pjob.reason = ''
                            for k in pjob.jobSpec.keys():
                                if k.endswith('ErrorDiag') and pjob.jobSpec[k]!='NULL':
                                    pjob.reason += '%s: %s, '%(k,str(pjob.jobSpec[k]))
                            #if job.backend.jobSpec['transExitCode'] != 'NULL':
                            pjob.reason += 'transExitCode: %s'%pjob.jobSpec['transExitCode']

                            if status.jobStatus in ['defined','unknown','assigned','waiting','activated','sent']:
                                logger.debug('Panda job %s %s' % (pjob.id, status.jobStatus))
                            elif status.jobStatus in ['starting','running','holding','transferring', 'merging']:
                                logger.debug('Panda job %s %s '% (pjob.id, status.jobStatus))
                            elif status.jobStatus in ['finished']:
                                logger.debug('Panda job %s %s '% (pjob.id, status.jobStatus))
                            elif status.jobStatus == 'failed':
                                logger.debug('Panda job %s %s '% (pjob.id, status.jobStatus))
                                # check for server side retry
                                if 'taskBufferErrorDiag' in pjob.jobSpec and pjob.jobSpec['taskBufferErrorDiag'].find("PandaID=") != -1:
                                    # grab the new panda ID
                                    newPandaID = long(pjob.jobSpec['taskBufferErrorDiag'].split("=")[1])
                                    pjob.id = newPandaID
                                    pjob.status = None
                                    pjob.url = 'http://panda.cern.ch/?job=%d'%newPandaID
                            elif status.jobStatus == 'cancelled' and pjob.status not in ['completed','failed']: # bug 67716
                                logger.debug('Panda job %s cancelled'%pjob.id)
                                if 'taskBufferErrorDiag' in pjob.jobSpec and "rebrokerage" in pjob.jobSpec['taskBufferErrorDiag']:
                                    newPandaID = checkForRebrokerage(pjob.jobSpec['taskBufferErrorDiag'])
                                    logger.warning("Subjob rebrokered by Panda server. Job %d moved to %d."%(pjob.id, newPandaID))
                                    pjob.id = newPandaID
                                    pjob.status = None
                            else:
                                logger.warning('Unexpected job status %s',status.jobStatus)

    def master_resubmit(self,jobs):
        '''Resubmit failed Jedi job'''
        from pandatools import Client

        jobIDs = {}
        for job in jobs: 
            jobIDs[job.backend.id] = job

        allJobIDs = jobIDs.keys()
        pandaJobIDs = {}
        for jID in allJobIDs:
            status, jediTaskDict = Client.getJediTaskDetails({'jediTaskID': jID},False,True,verbose=False)
            if status != 0:
                logger.error("Failed to get task details for %s" % jID)
                raise BackendError('Jedi','Return code %d retrieving job status information.' % status)

            # Retrieve job
            job = jobIDs[jediTaskDict['jediTaskID']]
       
            newJobsetID = -1 # get jobset
            retryJobs = [] # jspecs
            resubmittedJobs = [] # ganga jobs

            if jediTaskDict['status'] in ['failed', 'killed', 'cancelled', 'aborted', 'broken', 'finished' ]:
                retryJobs.append(job)
                resubmittedJobs.append(jID)
            #elif jediTaskDict['status'] == 'finished':
            #    pass
            else:
                logger.warning("Cannot resubmit. Jedi task %s is status %s." %(jID, jediTaskDict['status'] ))
                return False

            # submit
            if len(retryJobs)==0:
                logger.warning("No failed jobs to resubmit")
                return False
            
            status,out = Client.retryTask(jID, verbose=False)                                                               
            if status != 0:
                logger.error(status)
                logger.error(out)
                logger.error("Failed to retry JobID=%s" % jID)                                                                                         
                return False
            tmpStat,tmpDiag = out
            if not tmpStat:
                logger.error(tmpDiag)
                logger.error("Failed to retry JobID=%s" % jID)
                return False
            logger.info(tmpDiag)
       
            job.backend.status = None
            job.backend.jobSpec = {}
            job.updateStatus('submitted')

        logger.info('Resubmission successful')
        return True

    def master_setup_bulk_subjobs(self, jobs, jdefids):
           
        from GangaCore.GPIDev.Lib.Job.Job import Job
        master_job=self.getJobObject()
        for i in range(len(jdefids)):
            j=Job()
            j.copyFrom(master_job)
            j.splitter = None
            j.backend=Panda()
            j.backend.id = jdefids[i]
            j.id = i
            j.status = 'submitted'
            j.time.timenow('submitted')
            master_job.subjobs.append(j)
        return True

    def get_stats(self):
        fields = {
            'site':"self.jobSpec['computingSite']",
            'exitstatus':"self.jobSpec['transExitCode']",
            'outse':"self.jobSpec['destinationSE']",
            'jdltime':"''",
            'submittime':"int(time.mktime(time.strptime(self.jobSpec['creationTime'],'%Y-%m-%d %H:%M:%S')))",
            'starttime':"int(time.mktime(time.strptime(self.jobSpec['startTime'],'%Y-%m-%d %H:%M:%S')))",
            'stoptime':"int(time.mktime(time.strptime(self.jobSpec['endTime'],'%Y-%m-%d %H:%M:%S')))",
            'totalevents':"int(self.jobSpec['nEvents'])", 
            'nInputDataFiles':"int(self.jobSpec['nInputDataFiles'])", 
            'inputFileType':"int(self.jobSpec['inputFileType'])", 
            'inputFileProject':"self.jobSpec['inputFileProject']", 
            'inputFileBytes':"int(self.jobSpec['inputFileBytes'])", 
            'nOutputDataFiles':"int(self.jobSpec['nOutputDataFiles'])", 
            'outputFileBytes':"int(self.jobSpec['outputFileBytes'])", 
            'jobMetrics':"self.jobSpec['jobMetrics']", 
            'wallclock':"(int(time.mktime(time.strptime(self.jobSpec['endTime'],'%Y-%m-%d %H:%M:%S')))-int(time.mktime(time.strptime(self.jobSpec['startTime'],'%Y-%m-%d %H:%M:%S'))))",
            'percentcpu':"int(100*self.jobSpec['cpuConsumptionTime']/float(self.jobSpec['cpuConversion'])/(int(time.mktime(time.strptime(self.jobSpec['endTime'],'%Y-%m-%d %H:%M:%S')))-int(time.mktime(time.strptime(self.jobSpec['startTime'],'%Y-%m-%d %H:%M:%S')))))",
            'numfiles':'""',
            'gangatime1':'""',
            'gangatime2':'""',
            'gangatime3':'""',
            'gangatime4':'""',
            'gangatime5':'""',
            'pandatime1':"int(self.jobSpec['pilotTiming'].split('|')[0])",
            'pandatime2':"int(self.jobSpec['pilotTiming'].split('|')[1])",
            'pandatime3':"int(self.jobSpec['pilotTiming'].split('|')[2])",
            'pandatime4':"int(self.jobSpec['pilotTiming'].split('|')[3])",
            'pandatime5':"int(self.jobSpec['pilotTiming'].split('|')[4])",
            'NET_ETH_RX_PREATHENA':'""',
            'NET_ETH_RX_AFTERATHENA':'""'
            }
        stats = {}
        for k in fields.keys():
            try:
                stats[k] = eval(fields[k])
            except:
                pass
        return stats

    def getStateTime(self, status):
        """Obtains the timestamps for the 'running', 'completed', and 'failed' states.

           The __jobstatus__ file in the job's output directory is read to obtain the start and stop times of the job.
           These are converted into datetime objects and returned to the user.
        """

        import datetime
        checkstr=''
        if status == 'running':
            checkstr='startTime'
        elif status == 'completed':
            checkstr='endTime'
        elif status == 'failed':
            checkstr='endTime'

        if not checkstr:
            return None

        try:
            t = datetime.datetime(*(time.strptime(self.jobSpec[checkstr], "%Y-%m-%d %H:%M:%S")[0:6]))
        except:
            return None
        return t


    def timedetails(self):
        """Return all available timestamps from this backend.
        """

        import datetime
        r = self.getStateTime('running')
        c = self.getStateTime('completed')
        d = {'START' : r, 'STOP' : c}

        return d





'''Module for processing XMLSchema.
The object VTree holds a validated tree of xml.
The object Schema holds the parsed xml schema.

The Schema object is the main functional object.
The VTree object is a light-weight wrapper of the validated xml code.
The VTree can only be altered in a way which conforms to the schema.

Schema:

    Default xml can be generated from the schema: create_default(tag)
    Existing xml can be parsed/validated against the schema: parse(file)
    Most of the xml 1.0 standard is allowed/parsed
    Should not be altered once the schema has been parsed

    print the schema to see all valid types and tags
    query the constraints of a tag with Tag_constraints(tag)
    helper functions to find out what is allowed are prefixed with Tag_
    functions which print out information about the schema start with lower case
    
    inconsistent or erroneous schema will throw a NameError, TypeError or AttributeError 
    
VTree:
    
    Typical data object validated against the schema
    Usually created from a default from the Schema object,
    or by parsing existing xml from the Schema object
    Can only be altered in a way which conforms to the schema.
    
    Attributes for getting and setting of attribs, and getting of children
    are automatically generated.
    sum.an_attribute_that_is_defined_by_schema(value)
    is a shortcut for
    sum.attrib('an_attribute_that_is_defined_by_schema', value)
    
    sum.a_child_that_is_defined_by_schema(attrib,value)
    is a shortcut for
    sum.children('a_child_that_is_defined_by_schema', child)
    
    trying to change the VTree in a way which does not conform to the schema
    will throw a NameError, ValueError, TypeError or AttributeError 
'''

id = '$Id: schema.py,v 1.5 2009-10-07 13:51:18 rlambert Exp $'

__author__ = 'Rob Lambert'
__date__ = id.split()[3]
__version__ = id.split()[2]

try:
    from xml.etree import cElementTree as __ElementTree__
except ImportError:
    try:
        from xml.etree import ElementTree as __ElementTree__
    except ImportError:
        try:
            from etree import ElementTree as __ElementTree__
        except ImportError:
            import ElementTree as __ElementTree__
            #finally fail here if module cannot be found!

import os as __os__


class VTree(object):
    '''a validated tree object
    The object holds a reference to the schema used
    print a VTree to see its content
    It is unlikely this class will be created by the user,
    rather it should be returned from parsing a file or creating a default object
    using the schema.
    Trying to change the VTree in a way which does not conform to the schema
    will throw a NameError, ValueError, TypeError or AttributeError 
    '''
    #__name__='VTree'
    #__type__='VTree'
    def __init__(self,element,schema, mother=None, check=True):
        '''constructor.
        Unlikely to be called without using the schema object directly
        From an etree element, and a schema the validated object will be formed
        mother is a pointer to the mother element of the tree, to keep track of the level
        check signifies if a recursive check against the schema is required'''
    #    docstr='''
    #This particular VTree is for the tag '###TAG###'
    #which means it gets the extra methods:'''
        #the schema object is validated against
        self.__schema__=schema
        #the validated object
        self.__element__=element
        self.__mother__=mother
        if check:
            if not self.__schema__.__check__(element):
                raise ValueError("cannot validate element")
        #self.__doc__=self.__doc__+docstr.replace('###TAG###', self.tag())
        ##auto generate methods for defined children and attribs
        #for key in self.attrib().keys():
        #    try:
        #        dstr="self.###ATTRIB###".replace("###ATTRIB###",key)
        #        exec dstr
        #    except (NameError, AttributeError, ValueError, SyntaxError): pass
        #for child in self.__schema__.Tag_children(self.tag()):
        #    try:
        #        dstr="self.###ATTRIB###".replace("###ATTRIB###",child)
        #        exec dstr
        #    except (NameError, AttributeError, ValueError, SyntaxError): pass
        #except all pre-thought of exceptions here... to avoid errors in init!
    def __reimport__(self):
        global __os__
        global __ElementTree__
        
        try:
            from xml.etree import ElementTree as __ElementTree__
        except ImportError:
            try:
                from etree import ElementTree as __ElementTree__
            except ImportError:
                import ElementTree as __ElementTree__
            #finally fail here if module cannot be found!
        import os as __os__
        
    def __children__(self):
        '''list the existing children'''
        list=[]
        for c in self.__element__.getchildren():
            list.append(c.tag)
        return list
    def __repr__(self):
        '''how to print this object'''
        return self.__str__()
    def __str__(self):
        '''how to print this object'''
        ret= 'VTree-'+self.tag()+': '
        v=self.value()
        if v is not None:
            if type(v)==type([]):
                ret+=self.__schema__.__list2str__(self.value())
            else: ret+=str(v)
        if len(self.attrib()):
            ret+='\n attrib='+str(self.attrib())
        if len(self.__children__()):
            ret+='\n children='+str(self.__children__())
        return ret
    
    def __is__(self, ele, tag=None, attrib=None, value=None):
        '''internal finding function used by find and children
        the default, None, makes no requirement on the children.
        Attrib can be a single att, a list of required attribs, or a dictionary of attrib:value
        value only be a single value, what you expect the object to hold
        tag can be a single tag, or a list
        multiple tags are ORED.
        multiple attributes are ANDED
        '''        
        #check for tags
        if tag is not None:
            if type([]) == type(tag):
                if ele.tag not in tag:
                    return False
            else:
                if tag!=ele.tag:
                    return False
        #check for attributes
        if attrib is not None:
            #list of attributes which must exist
            if str(type([])) in str(type(attrib)):
                for att in attrib:
                    if att not in ele.attrib: return False
                
            #list of attribute:value pairs
            elif str(type({})) in str(type(attrib)):
                for att in attrib:
                    if att not in ele.attrib: return False
                    vc=self.__schema__.Tag_castValue(att,attrib[att])
                    vatt=self.__schema__.Tag_castValue(att, ele.attrib[att])
                    if vc is None or vatt is None: return False
                    if vc!=vatt: return False
            elif attrib not in ele.attrib: return False
        #check for value
        if value is not None:
            if not ele.text: return False
            try:
                value=self.__schema__.Tag_castValue(ele.tag,value)
            except (ValueError, TypeError, KeyError):
                return False
            if value!=self.__schema__.Tag_castValue(ele.tag,ele.text): return False
                
        return True
    
    #def __getattr__(self, name):
    #    '''append a get/set method with the name of the attribute'''
    #    docstr=""
    #    dstr='''class tmp_###ATTRIB###:
    #def __init__(self,V):
    #    self.V=V''' 
    #    type=None
    #    if name in self.attrib():
    #        docstr='''
    #    ###ATTRIB###(value=None), to get/set '###ATTRIB###' '''
    #        dstr=dstr+'''
    #def value(self, value=None):
    #    "shorcut to set/get ###ATTRIB###, sum.attrib('###ATTRIB###', value)" 
    #    return self.V.attrib('###ATTRIB###',value)'''
    #    elif name in self.__schema__.Tag_children(self.tag()):
    #        docstr='''
    #    ###ATTRIB###(attrib=None, value=None), to get a list of '###ATTRIB###' children'''
    #        dstr=dstr+'''
    #def value(self, attrib=None, value=None):
    #    "shorcut to get ###ATTRIB###, sum.children('###ATTRIB###', attrib, value)" 
    #    return self.V.children('###ATTRIB###',attrib,value)'''
    #    else:
    #        raise AttributeError, name + ' is not a valid attribute or child'
    #    
    #    docstr=docstr.replace('###ATTRIB###',name)
    #    dstr=dstr.replace('###ATTRIB###',name)
    #    #print dstr
    #    exec dstr
    #    dstr="self.###ATTRIB###=tmp_###ATTRIB###(self).value".replace('###ATTRIB###',name)
    #    #print dstr
    #    exec dstr
    #    dstr="ret=self.###ATTRIB###".replace('###ATTRIB###',name)
    #    #print dstr
    #    exec dstr
    #    #only increase the docstring if the process completes OK
    #    self.__doc__+=docstr
    #    return ret

    
        
    def __append_element__(self, child):
        '''internal method to append validated elements'''
        self.__element__.append(child.__element__)
        
        #set the level
        #if there's text there, ignore!
        if self.__element__.text is not None and self.__element__.text.strip()!='':
            return True
        #get the level from the mother
        if self.__element__.text is None:
            if (self.__mother__ is None or
                self.__mother__.__element__.text is None or
                self.__mother__.__element__.text.strip()!=''):
                self.__element__.text='\n\t'
                #self.__element__.tail='\n'
            else:
                self.__element__.text=self.__mother__.__element__.text+'\t'
                #self.__element__.tail=self.__mother__.__element__.text
        
        #if there's text there, ignore!
        #if self.__element__.text is None or self.__element__.text.strip()=='':
        self.__element__.getchildren()[-1].tail=str(self.__element__.text)[:-1]
        if len(self.__element__.getchildren())>1:
            if self.__element__.text is None or self.__element__.text.strip()=='':
                self.__element__.getchildren()[-2].tail=str(self.__element__.text)
        return True


    def __insert_element__(self,child,index):
        '''internal method to insert validated elements'''
        element_size=len(self.__element__.getchildren())

        if (element_size==0 or index>=element_size):
            return self.__append_element__(child)

        if (index<=(-element_size)):
            real_index=0
        elif (index<0):
            real_index=index + element_size
        else:
            real_index=index     
        self.__element__.insert(real_index,child.__element__)
        
        #set the level
        # now size==element_size+1
        if real_index!=element_size-1:
            self.__element__.getchildren()[real_index].tail=self.__element__.getchildren()[real_index+1].tail
        else:
            self.__element__.getchildren()[real_index].tail=self.__element__.getchildren()[real_index+1].tail[:-1]

        return True

    def __remove_element__(self,child):
        '''internal method to remove validated elements'''
        children=self.__element__.getchildren()
        element_size=len(children)
        if element_size==1:
            self.__element__.remove(child.__element__)
            # remove text if there were only \n's or \t's
            if self.__element__.text is not None and self.__element__.text.strip()=='':
                self.__element__.text=None
            return True
        elif element_size>1:
            if child.__element__==children[element_size-1]:
                # copy tail on the previous child
                children[-2].tail=children[-1].tail
            self.__element__.remove(child.__element__) 
            return True
        else:
            raise TypeError('This should never happen since child is supposed to belong to the children of the element')
            return False

    def test(self):
        '''tests that the object is OK'''
        try:
            if (__ElementTree__ is None or __os__ is None or
                self.__schema__ is None or self.__element__ is None):
                return False
            self.__element__.text
            return True
        except:
            return False
    
    def dump(self):
        '''dump the object to the screen'''
        print(self.xml())
    def xml(self):
        '''dump the object to an xml string'''
        if __os__ is None or __ElementTree__ is None:
            self.__reimport__()
            if __os__ is None or __ElementTree__ is None:
                raise ImportError("problem with modules!")
                return False #not needed
        return self.__schema__.header()+'\n'+__ElementTree__.tostring(
            self.__element__).replace(
            'ns0:',self.__schema__.__ns__.rstrip(':')+'i:').replace(
            ':ns0',':'+self.__schema__.__ns__.rstrip(':')+'i').replace(
            self.__schema__.__schemafile_long__,self.__schema__.__schemafile_short__)
    def write(self,outfile):
        ''' write xml to a file'''
        if outfile is None: return False
        if outfile=="": return False
        if __os__ is None or __ElementTree__ is None:
            self.__reimport__()
            if __os__ is None or __ElementTree__ is None:
                raise ImportError("problem with modules!")
                return False #not needed
        outfile=__os__.path.expanduser(__os__.path.expandvars(outfile))
        if not self.__schema__.Tag_isRoot(self.tag()):
            raise IOError("cannot output a file which doesn't have the root object")
            return False #not needed
        f=open(outfile,'w')
        if not f:
            raise IOError('Error opening file for writing'+str(outfile))
            return False
        f.write(self.xml())
        f.close()
        return True
    def constraints(self):
        '''get the constraints from the schema '''
        return self.__schema__.Tag_constraints(self.__element__.tag)
    def attrib(self, att=None, val=None):
        '''return a dictionary of the existing attributes
        this dictionary is not connected to the element
        if att is set, will return only the value of the given attribute
        if val is set, will set att to the given val'''
        #print all attribs
        if att is None:
            list={}
            for c in self.__element__.attrib:
                list[c]=self.__schema__.__cast_from_tag__(c,self.__element__.attrib[c])
            return list
        #print the given attrib
        if val is None:
            try:
                return self.__schema__.__cast_from_tag__(att,self.__element__.attrib[att])
            except KeyError:
                raise AttributeError(self.tag()+" has no attribute " + att)
                return None
        #set the given attrib
        if not self.__schema__.Tag_hasAttrib(self.tag(),att):
            raise AttributeError(self.tag()+" has no attribute " + att)
            return False
        
        #elif self.__schema__.Tag_canHaveValue(att,val):
        self.__element__.attrib[att]=str(self.__schema__.Tag_castValue(att,val))
        if self.__schema__.Tag_whitespace(att)=='collapse':
            self.__element__.attrib[att]=self.__element__.attrib[att].rstrip().lstrip()
        return True
        #else:
        #    raise ValueError, att+" cannot take a value " + str(val)
        #    return False
    def addto(self, mother, child):
        '''Add the child to the first mother'''
        return self.find(mother)[0].add(child)

    # ORDER IN SEQUENCES IS NOT CHECKED !!!
    def add(self, child,index=None):
        '''add a child to the tree'''
        name=''
        if 'VTree' in child.__str__():
            name=child.tag()
        elif 'str' in str(type(child)):
            name=child
        else:
            try:
                name=child.tag()
            except (AttributeError, ValueError, TypeError):
                raise TypeError('you can only add VTree objects, or use a string to add the default object')
                return False
        if name not in self.__schema__.Tag_children(self.tag()):
            raise TypeError('cannot add '+ name+ ' to ' + self.tag()+ ' as this child is not allowed in the schema,')
            return False
        if self.nChildren(name)==self.__schema__.Tag_nChild(self.tag(),name)[1]:
            raise TypeError('cannot add'+ name+ ' to ' + self.tag()+ ' as there are enough of this child already', name)
            return False
        if 'VTree' in child.__str__():
            if index is None:
                return self.__append_element__(child)
            else:
                return self.__insert_element__(child,index)
        else:
            if index is None:
                return self.__append_element__(self.__schema__.create_default(name))
            else:
                return self.__insert_element__(self.__schema__.create_default(name),index)


    def remove(self,child):
        '''remove a child from the tree'''
        if child.__element__ not in self.__element__.getchildren():
            raise TypeError('This object does not contain this child, the child cannot be removed')
            return False
        if self.nChildren(child.tag())==self.__schema__.Tag_nChild(self.tag(),child.tag())[0]:
            raise TypeError('cannot remove '+ child.tag()+ ' to ' + self.tag()+ ' as there will not be enough of this child')
            return False
        else:
            return self.__remove_element__(child)
        
    
    def value(self, val=None):
        '''return the existing value, or None
        a string is returned which is not connected to the element
        if val is given, will set the value to this'''
        #return the value
        #print 'in value using', self.__element__.text
        if val is None:
            if self.__element__.text is None:
                return None
            if len(self.__element__.text.rstrip().lstrip())==0:
                return None
            if self.__schema__.Tag_whitespace(self.tag())=='collapse':
                return self.__schema__.__cast_from_tag__(self.tag(),self.__element__.text.rstrip().lstrip())
            return self.__schema__.__cast_from_tag__(self.tag(),self.__element__.text)
        #set the value
        v=self.__schema__.Tag_castValue(self.tag(),val)
        #if self.__schema__.Tag_canHaveValue(self.tag(),val):
        #    v=self.__schema__.Tag_castValue(self.tag(),val)
        if str(type([]))==str(type(v)):
            v=self.__schema__.__list2str__(v)
        self.__element__.text=str(v)
        if self.__schema__.Tag_whitespace(self.tag())=='collapse':
            self.__element__.text=self.__element__.text.rstrip().lstrip()
        return True
        #else:
        #    raise ValueError, self.tag()+" cannot take a value " + str(val)
        #    return False
    def tag(self):
        '''return the existing tag'''
        return ''+self.__element__.tag
    def nChildren(self, tag=''):
        '''how many children of this type are there?'''
        if tag=='':
            return len(self.children())
        else:
            n=0
            for c in self.__children__():
                if c==tag:
                    n=n+1
        return n
    def mother(self):
        '''return the mother of this VTree element
        used to allow full tree navigation'''
        return self.__mother__
    def children(self, tag=None, attrib=None, value=None):
        '''return a list of the direct children with the given tags, attributes and values.
        the default, None, makes no requirement on the children.
        Attrib can be a list of required attribs, or a dictionary of attrib:value
        value only be a single value, what you expect the object to hold
        tag can be a single tag, or a list
        multiple tags are ORED.
        multiple attributes are ANDED
        '''
        it=self.__element__.getchildren()
        list=[]
        for child in it:
            if self.__is__(child,tag,attrib,value):
                list.append(VTree(child,self.__schema__,self,False))
        return list
    
    def find(self, tag=None, attrib=None, value=None):
        '''return a list of the elements with the given tags, attributes and values.
        from this level downwards.
        the default, None, makes no requirement on the daughters.
        Attrib can be a list of required attribs, or a dictionary of attrib:value
        value can be a single value, or a list of strings
        tag can be a single tag, or a list
        tags and values are ORED.
        attributes are ANDED'''
        it=None
        if tag is not None:
            if str(type([])) in str(type(tag)):
                it=self.__element__.getiterator()
            else:
                it=self.__element__.getiterator(tag)
        else:
            it=self.__element__.getiterator()
        list=[]
        for child in it:
            if self.__is__(child,tag,attrib,value):
                list.append(VTree(child,self.__schema__,self,False))
        return list
    
    def __clone_element__(self, ele):
        def_e=__ElementTree__.Element(ele.tag,ele.attrib)
        def_e.text=ele.text
        def_e.tail=ele.tail
        for c in ele.getchildren():
            def_e.append(self.__clone_element__(c))
        return def_e
    
    #def __internal_clone__(self):
    #    '''return a clone of this element, with the same schema'''
    #    e2=self.__clone_element__(self.__element__)
    #    self.__element__==e2
        
    def clone(self):
        '''return a clone of this element, with the same schema'''
        e2=self.__clone_element__(self.__element__)
        return VTree(e2,self.__schema__,None,False)

class Schema(object):
    '''details about the xml schema
    The default constructor should parse most schema
    VTrees are created by the create_default(), parse() and validate() methods
    print a parsed schema to see its content
    inconsistent or erroneous schema will throw a NameError, TypeError or AttributeError '''
    #cached default elements
    #__def_cache__={}
    def __init__(self, schemafile, ns='xs', root=''):
        '''constructor. 
        schemafile is the name of the file containing the schema, usually an xsd file
        If the namespace is not equal to xs change by setting the value of ns=
        if the root element is not the first/only defined element, set with option root='''
        self.__tree__=__ElementTree__.ElementTree()
        #that is the parsed schema
        self.__header__=''
        self.__schemafile_long__=__os__.path.expanduser(__os__.path.expandvars(schemafile))
        self.__schemafile_short__=schemafile
        self.__root__=root
        self.__rootattribs__={}
        self.__def_cache__={}
        self.__mother_cache__={}
        self.__child_cache__={}
        self.__attrib_cache__={}
        self.__type_cache__={}
        self.__enum_cache__={}
        self.__fixed_cache__={}
        self.__default_cache__={}
        self.__whitespace_cache__={}
        self.__union_cache__={}
        self.__list_cache__={}
        self.__seq_cache__={}
        self.__ns__=ns+':'
        self.__uri__="{http://www.w3.org/2001/XMLSchema}"
        #all known tags
        self.__tags__=set()
        self.__tagelement__={}
        #all known types
        self.__basetypes__=set(["integer",
                        "long",
                        "unsignedLong",
                        "double",
                        "string",
                        "normalizedString",
                        "boolean"])
        #types I can cast into
        self.__cast_types__=set([t.lower() for t in self.__basetypes__ ])
        #dictionary to remove namespace without string manipulation
        self.__type_remove_namespace__={}
        self.__func_cast__={
            'n' : lambda x: str(x),   #normalizedString
            's' : lambda x: str(x),   #string
            'f' : lambda x: float(x), #float
            'l' : lambda x: int(x),  #long
            'i' : lambda x: int(x),   #integer
            'd' : lambda x: float(x)  #double
            }
        self.__types__=set()
        self.__typelement__={}
        #all known attributes
        self.__attribs__=set()
        #type of that attribute
        self.__attribelement__={}
        #self.__attrib_rules__=[
        #                "default",
        #              r  "fixed",
        #                "minOccurs",
        #                "maxOccurs",
        #                "use" ]
        #all known rules
        #self.__rules__=["enumeration",
        #                "whiteSpace"]
        if self.__schemafile_short__:
            pf=open(self.__schemafile_long__)
            self.__header__=pf.readlines()[0]
            pf.close()
            self.__parseschema__(self.__schemafile_long__)
            self.__isconsistent__()
            #if self.__root__=='' and len(self.__tags__)>0:
            #    self.__root__=self.__tags__[0]
            if self.__root__ not in self.__tags__:
                raise TypeError('root of schema not found '+self.__root__)
                return self.__init__()
            self.__rootattribs__[
                    "xmlns:"+self.__ns__.split(':')[0]+"i"
                        ]=(
                           self.__uri__.split('}')[0].split('{')[-1]
                           +"-instance"
                           )
            self.__rootattribs__[
                    self.__ns__.split(':')[0]+"i:noNamespaceSchemaLocation"
                        ]=(
                           self.__schemafile_long__
                           )
            self.__rootattribs__[
                self.__uri__.split('}')[0]
                +"-instance}"
                +"noNamespaceSchemaLocation"
                        ]=(
                           self.__schemafile_long__
                           )
            
    # ORDER IN SEQUENCES IS NOT CHECKED !!!
    def __check__(self,element):
        '''internal method to check an element conforms to the schema'''
        #check tag
        if element.tag not in self.tags():
            raise NameError('element '+ element.tag+ ' does not exist in the schema')
            return False
        #check value
        if element.text:
            if len(element.text.rstrip().lstrip()):
                if not self.Tag_canHaveValue(element.tag, element.text):
                    raise ValueError('element '+ element.tag+ ' has the wrong entry type for the schema')
                    return False
        #check attribs
        for att,val in element.items():
            #check the root attributes
            #if self.Tag_isRoot(element.tag):
            #    for att in self.__rootattribs__.keys():
            #        if att not in element.attrib.keys():
            #            print 'the root element', element.tag, 'must have the attrib', att
            #            return False
            #        if self.__rootattribs__[att]!=element.attrib[att]:
            #           print 'the root attribute', att, 'must have the value', self.__rootattribs__[att]
            #            return False
            if not self.Tag_hasAttrib(element.tag,att):
                raise AttributeError('element '+ element.tag+ ' cannot have attribute '+att +' in the schema')
                return False
            if self.Tag_isRoot(element.tag) and 'noNamespaceSchemaLocation' in att:
                if element.attrib[att].split('/')[-1]!=self.__schemafile_long__.split('/')[-1]:
                    raise AttributeError('root element '+ element.tag+
                                           ' must be from the same schema!! '+
                                            ' attribute '+att +' is '+ val+
                                            ' versus '+self.__schemafile_long__
                                            )
                    return False
            elif not self.Tag_canHaveValue(att, val):
                raise AttributeError('element '+ element.tag + 
                                       ' cannot have attribute ' + att +
                                       ' with value '+ val+
                                       ' in the schema'
                                       )
                return False
        #check required attribs
        for att in self.Tag_attribs(element.tag):
            if self.Tag_isAttribRequired(element.tag, att):
                if att not in element.attrib:
                    raise AttributeError( 'element '+ element.tag+ 
                                            ' must have attribute '+att +' in the schema'
                                            )  
                    return False
        
        #check children
        kiddic={}
        for child in element.getchildren():
            if child.tag in kiddic:
                kiddic[child.tag]=kiddic[child.tag]+1
            else:
                kiddic[child.tag]=1
            if child.tag not in self.Tag_children(element.tag):
                raise AttributeError( 'element '+ element.tag+ 
                                        'cannot have child '+child +' in the schema'
                                        )  
                return False
            if not self.__check__(child):
                return False
        #check number of children
        for child in self.Tag_children(element.tag):
            if self.Tag_nChild(element.tag, child)[0]>0:
                #print child, kiddic, element.tag
                try:
                    if self.Tag_nChild(element.tag, child)[0]>kiddic[child]:
                        raise AttributeError( 'element '+ element.tag+ 
                                                ' has not enough copies of '+child +
                                                ' for the schema'
                                                )
                        return False
                except KeyError:
                    raise AttributeError( 'element '+ element.tag+ 
                                            ' requires child '+child +
                                            ' for the schema'
                                            )
                    
            if self.Tag_nChild(element.tag, child)[1]>0:
                try:
                    if self.Tag_nChild(element.tag, child)[1]<kiddic[child]:
                        print('element', element.tag, 'has too many copies of',child ,'for the schema')
                        return False
                except KeyError:
                    pass
        return True
    
    def __str__(self):
        '''what to print to the screen'''
        return 'tags='+str(self.__tags__)+'\ntypes='+str(self.__types__)+'\nattribs='+str(self.__attribs__)+'\nroot='+self.__root__+':'+str(self.__rootattribs__)

    def __parseschema__(self, parsefile):
        '''internal method to parse a file into the schema'''
        #parse an existing schema
        if __os__.path.exists(parsefile):
            self.__tree__=__ElementTree__.parse(parsefile)
            rt=self.__tree__.getroot()
            if rt:
                self.__uri__=rt.tag[:rt.tag.find('schema')]
                for i in list(self.__basetypes__):
                    #add namespace to basic types
                    self.__basetypes__.add(
                        self.__ns__+i)
                    self.__basetypes__.add(self.__uri__+i)
                    self.__basetypes__.remove(i)
                    self.__type_remove_namespace__[self.__uri__+i]=i.lower()
                    self.__type_remove_namespace__[self.__ns__+i]=i.lower()
                    
                    
                for e in rt.getiterator( self.__uri__+"element"):
                    try:
                        self.__tags__.add( e.attrib['name'])
                        #print 'adding tagelement'
                        self.__tagelement__[e.attrib['name']]=e
                        #print 'added tagelement'
                        #root by default is the first thing in the list
                        if self.__root__=='':
                            self.__root__=e.attrib['name']
                    except KeyError: pass
                for e in rt.getiterator( self.__uri__+"simpleType"):
                    try:
                        self.__types__.add( e.attrib['name'])
                        #print 'adding typelement'
                        self.__typelement__[e.attrib['name']]=e
                        #print 'added typelement'
                    except KeyError: pass
                for e in rt.getiterator( self.__uri__+"complexType"):
                    try:
                        self.__types__.add( e.attrib['name'])
                        #print 'adding typelement'
                        self.__typelement__[e.attrib['name']]=e
                        #print 'added typelement'
                    except KeyError: pass
                for e in rt.getiterator( self.__uri__+"attribute"):
                    try:
                        self.__attribs__.add( e.attrib['name'])
                        #print 'adding attribelement'
                        self.__attribelement__[e.attrib['name']]=e
                        #print 'added attribelement'
                    except KeyError: pass
                return True
        return False
    
    def __isconsistent__(self):
        '''internal method to check the file is a consistent schema'''
        #check schema is self-consistent
        #all schema entries must have name or ref from list
        #all entries must have type or base from list
        rt=self.__tree__.getroot()
        if rt:
            for e in rt.getiterator():
                if self.__uri__ not in e.tag:
                    __ElementTree__.dump(e)
                    raise NameError("unknown element, without namespace "+ e.tag)
                    return False
                aname=self.__ename__(e)
                #print aname
                if not aname: continue
                if (aname not in self.__tags__ and
                    aname not in self.__attribs__ and
                    aname not in self.__types__ and
                    aname not in self.__basetypes__ ):
                    __ElementTree__.dump(e)
                    raise NameError("unknown element name "+ e.tag)
                    return False
                atype=self.__etype__(e)
                if not atype: continue
                if str(type([])) not in str(type(atype)):
                    atype=[atype]
                for aatype in atype:
                    #print aatype
                    if (aatype not in self.__types__
                        and aatype not in self.__basetypes__):
                        __ElementTree__.dump(e)
                        raise TypeError("unknown element type "+ aatype)
                        return False
        #print 'schema is self-consistent'
        return True
    
    def __ename__(self, e):
        '''internal method to return the name of e'''
        try:
            aname=e.attrib['name']
            return aname
        except KeyError:
            try:
                aname=e.attrib['ref']
                return aname
            except KeyError:
                pass
            
            return None
        
    def __etype__(self, e):
        '''internal method to return the type or list of types of e'''
        try:
            atype=e.attrib['type']
            return atype
        except KeyError:
            #print 'not simple type, looking at children'
            try:
                for ec in e.getiterator():
                    if 'restriction' in ec.tag or 'extension' in ec.tag:
                        #print 'found restriction'
                        try:
                            atype=ec.attrib['base']
                            return atype
                        except KeyError: continue
                    elif 'sequence' in ec.tag:
                        #print 'found sequence'
                        atype=[]
                        for ecc in ec.getiterator():
                            if 'element' in ecc.tag:
                                atype.append(self.__etype__(ecc))
                        #print 'returning', atype
                        return atype
                    elif 'list' in  ec.tag:
                        #print 'found list'
                        try:
                            atype=ec.attrib['itemType']
                            return [atype]
                        except KeyError: continue
                    elif 'union' in  ec.tag:
                        #print 'found union'
                        try:
                            atype=ec.attrib['memberTypes']
                            return atype.split(' ')
                        except KeyError: continue
            except (KeyError, ValueError): pass
        return None
    
    def __checkcast__(self, atype,test):
        '''internal method, can this value be cast to the type of the tag?'''
        #print 'checking cast of', test ,'to', atype
        try:
            self.__cast__(atype,test)
            return True
        except (ValueError, TypeError):
            return False
    
    def __cast__(self, atype,test):
        '''internal method cast to the type of the tag'''
        #print 'cast of', test ,'to', atype
        btype=atype
        if btype in self.__type_remove_namespace__:
            btype=self.__type_remove_namespace__[atype]
        else:
            btype=btype.lower()
        
        #print 'moving to', btype
        
        a=False
        #should I make it absolute?
        if 'unsigned'==btype[:8]:
            a=True
            btype=btype[8:]
        
        if btype in self.__cast_types__:
            
            if 'b'==btype[0]:
                #print 'checking bool'
                if type('')==type(test):
                    if test.lower()=='true':
                        return True
                    elif test.lower()=='false':
                        return False
                    return bool(float(test))
                return bool(float(test)!=0)
            else:
                v=self.__func_cast__[btype[0]](test)
                if a:
                    v=abs(v)
                return v
        
        raise TypeError('do not no how to convert to type '+ atype)
        return None
    
    def __list2str__(self,list):
        '''internal method cast a list to a string separated by spaces'''
        rets=''
        for l in list:
            rets+=str(l)+' '
        return rets.rstrip()
    def __constraint__(self, tag, constr):
        '''internal method get the value for this constraint as a string'''
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                return None
            raise NameError( str(tag)+' tag is not in the schema' )
        try:
            return ele.attrib[constr]
        except KeyError: pass
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.__constraint__(atype,constr)
        
    def __default__(self, tag):
        '''internal method get the default value for this tag/attribute as a string'''
        return self.__constraint__(tag, 'default')
    
    def __fixed__(self,tag):
        '''internal method to return fixed value for this tag as a string'''
        return self.__constraint__(tag, 'fixed')
    
    def __enumeration__(self, tag):
        '''internal method to return the Enumerated values for this tag/attribute as strings'''
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                return None
            raise NameError(str(tag)+' tag is not in the schema' )
        enums=[]
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'enumeration'):
                try:
                    enums.append(ec.attrib['value'])
                except KeyError: pass
            if (len(enums)):
                return enums
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.__enumeration__(atype)
        return None
    
    def __getele__(self,tag, atag=True, aatt=True,atype=True):
        '''internal method to return the element corresponding to a tag'''
        ele=None
        try:
            if atag and tag in self.__tags__:
                ele=self.__tagelement__[tag]
            elif aatt and tag in self.__attribs__:
                ele=self.__attribelement__[tag]
            elif atype and tag in self.__types__:
                ele=self.__typelement__[tag]
        except (KeyError,AttributeError): pass
        return ele 
        
    def __compound__(self,tag,compound):
        '''sequence, list, union?'''
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                return False
            raise NameError(str(tag)+' tag is not in the schema' )
        for e in ele.getiterator(self.__ns__+compound):
            return True
        for e in ele.getiterator(self.__uri__+compound):
            #print 'iterated'
            return True
        for atype in self.__types__:
            if self.__etype__(ele)==atype:
                #print 'is of type', atype
                return self.__compound__(atype,compound)
        return False
        
    def __hasConstraint__(self,tag,constr):
        '''internal method to check for default, fixed'''
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                return False
            raise NameError( str(tag)+' tag is not in the schema' )
        try:
            ele.attrib[constr]
            return True
        except KeyError: pass
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.__hasConstraint__(atype,constr)
        return False

    def __fast_parse__(self,xmlfile):
        '''Parse without checking'''
        xmlfile=__os__.path.expanduser(__os__.path.expandvars(xmlfile))
        if not __os__.path.exists(xmlfile):
            raise IOError( 'file does not exist '+str(xmlfile))
            return None
        
        self.__tree__=__ElementTree__.parse(xmlfile)
        rt=self.__tree__.getroot()
        if not rt or not self.Tag_isRoot(rt.tag):
            raise TypeError( 'This file does not have the root of the schema')
            return None
        return VTree(rt,self,None,False)
                                                                                                
    def parse(self,xmlfile):
        '''parse an xml document and validate against this schema'''
        thetree=self.__fast_parse__(xmlfile)
        if not self.__check__(thetree.__element__):
            raise TypeError( 'This file could not be validated against the schema')
            return None
        #print 'file', xmlfile, 'sucessfully validated against the schema'
        return thetree
    
    def validate(self,xmlfile):
        '''parse an xml document and validate against this schema'''
        return self.parse(xmlfile)
    
    def create_default(self, tag, level=0):
        '''return the default validated object from the schema'''
        if tag not in self.__tags__:
            raise NameError( 'This tag, ' + tag + ' is not in the schema')
            return None
        
        if (level==0) and (tag in self.__def_cache__):
            #return a clone
            #self.__def_cache__[tag].__internal_clone__()
            return self.__def_cache__[tag].clone()
        
        def_e=__ElementTree__.Element(tag)
        #fill attributes
        for att in self.Tag_attribs(tag):
            #print 'setting default for', att
            if self.Tag_isAttribRequired(tag,att):
                if self.Tag_isFixed(att):
                    def_e.attrib[att]=self.__fixed__(att)
                elif self.Tag_hasDefault(att):
                    def_e.attrib[att]=self.__default__(att)
                elif self.Tag_hasEnumeration(att):
                    def_e.attrib[att]=self.__enumeration__(att)[0]
                else:
                    def_e.attrib[att]=''
                #print def_e.attrib[att]
        #fill children, iterate    
        for child in self.Tag_children(tag):
            cn=0
            if self.Tag_nChild(tag,child)[0]>cn:
                def_e.append(self.create_default(child,level+1).__element__)
                cn+=1
        if len(def_e.getchildren()):
            #add new line
            if def_e.text: def_e.text+='\n'
            else:
                def_e.text='\n'
            #add justification of first child
            for i in range(level+1):
                def_e.text+='\t'
            #adjust justification of last child
            def_e.getchildren()[-1].tail='\n'
            for i in range(level):
                def_e.getchildren()[-1].tail+='\t'
                    
        #fill value
        if self.__fixed__(tag) is not None:
            def_e.text=self.__fixed__(tag)
        elif self.__default__(tag) is not None:
            def_e.text=self.__default__(tag)
        elif self.__enumeration__(tag) is not None:
            def_e.text=self.__enumeration__(tag)[0]
        #tail value
        def_e.tail='\n'
        for i in range(level):
            def_e.tail+='\t'
        #check if it is root
        if self.Tag_isRoot(tag):
            def_e.attrib[list(self.__rootattribs__.keys())[1]]=self.__rootattribs__[list(self.__rootattribs__.keys())[1]]
            #for key in self.__rootattribs__.keys()[:2]:
            #    def_e.attrib[key]=self.__rootattribs__[key]
        
        retree=VTree(def_e,self,None,False)
        if (level==0): self.__def_cache__[tag]=retree
        
        return retree.clone()
    
    def header(self):
        '''return the xml header'''
        return self.__header__
    def root(self):
        '''return the xml root'''
        return self.__root__
    def setroot(self,tag):
        '''set the xml root'''
        if tag in self.__tags__:
            self.__root__=tag
            return True
        else: 
            raise NameError( 'This tag, ' + tag + ' is not in the schema')
            return False
    def dump(self):
        '''dump the object to an xml string'''
        print(self.xml())
    def xml(self):
        '''dump the object to an xml string'''
        return self.__header__+'\n'+__ElementTree__.tostring(self.__tree__.getroot()).replace(
            'ns0:',self.__ns__).replace(
            ':ns0',':'+self.__ns__.rstrip(':')
            ).replace(
                      self.__schemafile_long__,self.__schemafile_short__)
    def types(self):
        '''list of defined types'''
        return set(self.__types__)
    def attribs(self):
        '''list of defined attributes'''
        return set(self.__attribs__)
    def tags(self):
        '''list of defined tags'''
        return set(self.__tags__)
    def Tag_isRoot(self,tag):
        '''return true if the tag is the xml root'''
        return tag==self.__root__
    def Tag_constraints(self, tag):
        '''print schema constraints for tag'''
        if (tag not in self.__tags__ and
            tag not in self.__attribs__ and
            tag not in self.__types__):
            return {}
        dets={}
        dets['ValueTypes']=self.Tag_valueTypes(tag)
        dets['Enumeration']=self.Tag_hasEnumeration(tag)
        dets['Fixed']=self.Tag_fixed(tag)
        dets['Default']=self.Tag_default(tag)
        dets['Atts']=self.Tag_attribs(tag)
        kiddic={}
        for c in self.Tag_children(tag):
            kiddic[c]=self.Tag_nChild(tag,c)
        dets['Children']=kiddic
        return dets

    #helper functions to check the schema objects

    def Tag_canHaveValue(self, tag,val):
        '''can I set this value to the tag?'''
        try:
            self.Tag_castValue(tag,val)
            return True
        except (ValueError, TypeError):
            return False
        

    def __cast_from_tag__(self, tag, val):
        '''internal method for casting directly'''
        if tag in self.__rootattribs__:
            try:
                return str(val)
            except ValueError:
                raise ValueError( 'This root tag '+tag+' must be a string ')
                return None
        if (tag not in self.__tags__ and
            tag not in self.__attribs__ and
            tag not in self.__types__):
            raise NameError( 'This tag '+tag+' does not exist in the schema ')
            return None
        
        types=self.Tag_valueTypes(tag)
        if types is None:
            raise TypeError( 'No types defined for '+tag)
            #print 'no types found!'
            return None

        ret=None
        if self.Tag_isList(tag):
            #print 'recognised list'
            if type(types)==type([]):
                types=types[0]
            nval=[]
            #print 'casting', val, 'type', str(type(val))
            if type('')==type(val):
                #print 'splitting'
                nval=val.split(' ')
            elif type(val)==type([]):
                nval=val
            else: nval=[val]
            #print 'casting', val 
            ret=[]
            for aval in nval:
                try:
                    ret.append(self.__cast__(types,aval))
                except ValueError: 
                    raise ValueError( 'This list does not accept '+str(type(aval))+' values')
        
        elif self.Tag_isUnion(tag):
            #print 'recognised union'
            if types!=str(type([])):
                types=[types]
            for atype in types:
                try:
                    r1=self.__cast__(atype,aval)
                    ret=r1
                    break
                except ValueError:
                    continue
        else:
            #print 'basic type'
            if type(types)==type([]):
                types=types[0]
            ret=self.__cast__(types,val)
        
        if ret is None:
            raise ValueError( 'This tag does not accept '+str(type(aval))+' values')
        return ret
    
    def Tag_castValue(self, tag,val):
        '''can I set this value to the tag?'''
        ret=self.__cast_from_tag__(tag,val)
        
        f=self.Tag_fixed(tag)
        if f is not None:
            #print 'recognised fixed'
            if ret != f:
                raise ValueError( 'This value is fixed, and not equal to this one '+str(ret))
        e=self.Tag_enumeration(tag)
        if e is not None and len(e)>0:
            #print 'recognised enum'
            if ret not in e:
                raise ValueError( 'This value is not one of the enumerated list '+str(ret))
            
        if ret is None:
            raise ValueError( 'This tag '+tag+' does not accept '+str(type(val))+' values'    )
        else:
            return ret
    
    
    def Tag_isSequence(self, tag):
        '''does this tag define a sequence?'''
        if tag in self.__seq_cache__:
            return self.__seq_cache__[tag]
        ise=self.__compound__(tag,'sequence')
        self.__seq_cache__[tag]=ise
        return ise
    
    def Tag_isList(self, tag):
        '''does this tag take a list?'''
        if tag in self.__list_cache__:
            return self.__list_cache__[tag]
        ile=self.__compound__(tag,'list')
        self.__list_cache__[tag]=ile
        return ile
    
    
    def Tag_isUnion(self, tag):
        '''does this tag take a union?'''
        if tag in self.__union_cache__:
            return self.__union_cache__[tag]
        iu=self.__compound__(tag,'union')
        self.__union_cache__[tag]=iu
        return iu
    
    def Tag_hasDefault(self,tag):
        '''is there a default value for this tag/attribute?'''
        return self.__hasConstraint__(tag,'default')
        
    def Tag_default(self, tag):
        '''get the default value for this tag/attribute'''
        if tag in self.__default_cache__:
            return self.__default_cache__[tag]
        defa=self.__default__(tag)
        if defa is None:
            self.__default_cache__[tag]=None
            return None
        defa=self.Tag_castValue(tag,defa)
        self.__default_cache__[tag]=defa
        return defa
    
    def Tag_isFixed(self, tag):
        '''is there a fixed value for this tag/attribute?'''
        if tag in self.__fixed_cache__:
            if self.__fixed_cache__[tag] is not None:
                return True
            return False
        return self.__hasConstraint__(tag,'fixed')

    def Tag_fixed(self, tag):
        '''what is the fixed value for this tag/attribute?'''
        if tag in self.__fixed_cache__:
            return self.__fixed_cache__[tag]
        fix=self.__fixed__(tag)
        if fix is None:
            self.__fixed_cache__[tag]=None
            return None
        fix=self.Tag_castValue(tag,fix)
        self.__fixed_cache__[tag]=fix
        return fix
    
    def Tag_hasEnumeration(self, tag):
        '''are there Enumerated values for this tag/attribute?'''
        if tag in self.__enum_cache__:
            if self.__enum_cache__[tag] is not None:
                if len(self.__enum_cache__[tag]) > 0:
                    return True
            return False
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                return False
            raise NameError( str(tag)+' tag is not in the schema' )
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'enumeration'):
                return True
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.Tag_hasEnumeration(atype)
        return False
    
    def Tag_enumeration(self, tag):
        '''return the Enumerated values for this tag/attribute?'''
        if tag in self.__enum_cache__:
            return self.__enum_cache__[tag]
        enum=self.__enumeration__(tag)
        if enum is None:
            self.__enum_cache__[tag]=None
            return None
        enums=[]
        for e in enum:
            enums.append(self.__cast_from_tag__(tag,e))
        self.__enum_cache__[tag]=enums
        return enums
    
    def Tag_valueTypes(self,tag):
        '''what are the allowed value types for this tag/attribute?'''
        if tag in self.__type_cache__:
            return self.__type_cache__[tag]
##         if self.Tag_isSequence(tag):
##             self.__type_cache__[tag]=None
##             return None
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                self.__type_cache__[tag]=tag
                return tag
            raise NameError( str(tag)+' tag is not in the schema' )
        atype=self.__etype__(ele)
        if str(type([])) in str(type(atype)):
            #print 'a list or so'
            valuetypes=[]
            for aatype in atype:
                valuetypes.append(self.Tag_valueTypes(aatype))
            self.__type_cache__[tag]=valuetypes
            return valuetypes
        if atype in self.__basetypes__:
            self.__type_cache__[tag]=atype
            return atype
        for ttype in self.__types__:
            if atype==ttype:#ele.attrib['type']==ttype:
                #print 'ready to go to', ttype
                nt=self.Tag_valueTypes(atype)
                self.__type_cache__[tag]=nt
                return nt
        self.__type_cache__[tag]=None
        return None
    
    def Tag_isAttribRequired(self,tag,att):
        '''is this attribute required?'''
        try:
            if self.Tag_isRoot(tag) and att in self.__rootattribs__:
                return( att==list(self.__rootattribs__.keys())[0])
            
            ele=self.__attribelement__[att]
            if ele.attrib['use']=='required':
                return True
        except KeyError: pass
        
        ele=self.__getele__(tag,True,False,True)
        if ele is None:
            if tag in self.__basetypes__:
                return False
            raise NameError( str(tag)+' tag is not in the schema' )
        #print 'looking for tag'
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'attribute'):
                for find in ['name','ref']:
                    try:
                        if ec.attrib[find]==att:
                            if ec.attrib['use']=='required':
                                return True
                    except KeyError: pass
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.Tag_isAttribRequired(atype,att)
        return False
    
    def Tag_whitespace(self,tag):
        '''how is whitespace handled for this object?'''
        if tag in self.__whitespace_cache__:
            return self.__whitespace_cache__[tag]
        ele=self.__getele__(tag)
        if ele is None:
            if tag in self.__basetypes__:
                self.__whitespace_cache__[tag]=None
                return None
            raise NameError( str(tag)+' tag is not in the schema' )
        #print 'looking for tag'
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'whiteSpace'):
                #print 'iterating'
                try:
                    self.__whitespace_cache__[tag]=ec.attrib['value']
                    return ec.attrib['value']
                except KeyError: pass
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                aw=self.Tag_whitespace(atype)
                self.__whitespace_cache__[tag]=aw
                return aw
        self.__whitespace_cache__[tag]=None
        return None
    
    def Tag_hasAttrib(self,tag,att):
        '''is this an attribute of the tag?'''
        if self.Tag_isRoot(tag) and att in self.__rootattribs__:
            return True
        if tag in self.__attrib_cache__:
            return (att in self.__attrib_cache__[tag])
        if att not in self.__attribs__:
            return False
        
        ele=self.__getele__(tag,True,False,True)
        if ele is None:
            if tag in self.__basetypes__:
                return False
            raise NameError( str(tag)+' tag is not in the schema' )
        
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'attribute'):
                for find in ['name','ref']:
                    #print 'iterating'
                    try:
                        if ec.attrib[find]==att:
                            return True
                    except KeyError: pass
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.Tag_hasAttrib(atype,att)
        
        return False
    
    def Tag_attribs(self,tag):
        '''list all attributes for the tag'''
        if tag in self.__attrib_cache__:
            return self.__attrib_cache__[tag]
        
        ele=self.__getele__(tag,True,False,True)
        if ele is None:
            if tag in self.__basetypes__:
                self.__attrib_cache__[tag]=None
                return None
            raise NameError( str(tag)+' tag is not in the schema' )
        #print 'looking for tag'
        
        atts=[]
        #if self.Tag_isRoot(tag):
        #    atts=self.__rootattribs__.keys
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'attribute'):
                for find in ['name','ref']:
                    try:
                        atts.append( ec.attrib[find])
                        #print 'appended', atts
                    except KeyError: pass
        for atype in self.__types__:
            #print 'looking at daughter types'
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                #print 'found daughter types'
                atts+=self.Tag_attribs(atype)
                #print 'appended', atts
        self.__attrib_cache__[tag]=atts
        return atts
    
    def Tag_children(self, tag):
        '''list all possible children for the tag'''
        if tag in self.__child_cache__:
            return self.__child_cache__[tag]

        #if not a sequence there are no children
        if not self.Tag_isSequence(tag):
            self.__child_cache__[tag]=[]
            return []
        #retreive the element based on the tag
        ele=self.__getele__(tag,True,False,True)
        if ele is None:
            self.__child_cache__[tag]=[]
            return []
        child=[]
        #I iterate over everything looking for elements
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'element'):
                if 'name' in ec.attrib:
                    if ec.attrib['name']==tag:
                        continue
                    child.append(ec.attrib['name'])
            if len(child):
                self.__child_cache__[tag]=child
                return child
        #then I go to the type if the tag doesn't define it
        et=self.__etype__(ele)
        if et in self.__types__:
            return self.Tag_children(et)
        
        self.__child_cache__[tag]=child
        return child
        
    def Tag_mothers(self, tag):
        '''list all possible mothers for the tag'''
        if tag in self.__mother_cache__:
            return self.__mother_cache__[tag]
        mothers=[]
        if self.Tag_isRoot(tag): return None
        if tag not in self.__tags__:
            raise NameError( str(tag)+' is not a valid tag in the schema')
            return None
        for mother in self.__tags__:
            if tag in self.Tag_children(mother):
                mothers.append(mother)
        self.__mother_cache__[tag]=mothers
        return mothers
    
    def Tag_nChild(self, tag, child):
        '''give the [min,max] number of children of this type for this tag'''
        if not self.Tag_isSequence(tag):
            return [0,0]
        if child not in self.__tags__:
            return [0,0]
        ele=self.__getele__(tag,True,False,True)
        if ele is None:
            raise NameError( str(tag)+' tag is not allowed children' )
        for pref in [self.__ns__,self.__uri__]:
            for ec in ele.getiterator(pref+'element'):
                try:
                    if ec.attrib['name']==child:
                        #print 'found child'
                        min=1;
                        max=1;
                        try: 
                            if ec.attrib['minOccurs']=='unbounded':
                                min=-1
                            else:
                                min=int(ec.attrib['minOccurs'])
                        except (KeyError,ValueError): pass
                        try:
                            if ec.attrib['maxOccurs']=='unbounded':
                                max=-1
                            else:
                                max=int(ec.attrib['maxOccurs'])
                        except (KeyError,ValueError): pass
                        return [ min, max]
                except KeyError: continue
        for atype in self.__types__:
            if self.__etype__(ele)==atype:#ele.attrib['type']==atype:
                return self.Tag_nChild(atype,child)
        return [0,0]
        



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
from GangaLHCb.Lib.LHCbDataset.BKQuery import BKQuery, BKQueryDict
from GangaCore.GPIDev.Schema import Schema, Version, ComponentItem, SimpleItem

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class BKTestQuery(BKQuery):
    ##     schema = {}
    ##     docstr = 'Bookkeeping query path (type dependent)'
    ##     schema['path'] = SimpleItem(defvalue='' ,doc=docstr)
    ##     docstr = 'Start date string yyyy-mm-dd (only works for type="RunsByDate")'
    ##     schema['startDate'] = SimpleItem(defvalue='' ,doc=docstr)
    ##     docstr = 'End date string yyyy-mm-dd (only works for type="RunsByDate")'
    ##     schema['endDate'] = SimpleItem(defvalue='' ,doc=docstr)
    ##     docstr = 'Data quality flag (string or list of strings).'
    # schema['dqflag'] = SimpleItem(defvalue='All',typelist=['str','list'],
    # doc=docstr)
    ##     docstr = 'Type of query (Path, RunsByDate, Run, Production)'
    ##     schema['type'] = SimpleItem(defvalue='Path',doc=docstr)
    # docstr = 'Selection criteria: Runs, ProcessedRuns, NotProcessed (only \
    # works for type="RunsByDate")'
    ##     schema['selection'] = SimpleItem(defvalue='',doc=docstr)
    _schema = BKQuery._schema.inherit_copy()
    _schema.datadict['dataset'] = ComponentItem(
        'datasets', defvalue=None, optional=1, load_default=False, doc='dataset', hidden=0)
    _schema.datadict['fulldataset'] = ComponentItem(
        'datasets', defvalue=None, optional=1, load_default=False, doc='dataset', hidden=1)
    _schema.datadict['fulldatasetptr'] = SimpleItem(
        defvalue=0, optional=0, load_default=True, doc='dataset position pointer', hidden=1, typeList=['int'])
    _schema.datadict['filesToRelease'] = SimpleItem(
        defvalue=3, optional=0, load_default=True, doc='number of files to release at a time', hidden=0, typeList=['int'])
    _category = 'query'
    _name = "BKTestQuery"
    _exportmethods = BKQuery._exportmethods
    _exportmethods += ['removeData']

    def getDataset(self):
        if self.fulldataset is None:
            self.fulldataset = LHCbDataset(super(BKTestQuery, self).getDataset().files)
        if self.dataset is None:
            self.dataset = LHCbDataset(self.fulldataset.files[:self.filesToRelease])
            self.fulldatasetptr = self.filesToRelease
        else:
            self.dataset.files += self.fulldataset.files[
                self.fulldatasetptr:self.fulldatasetptr + self.filesToRelease]
            self.fulldatasetptr += self.filesToRelease
        return self.dataset

    def removeData(self):
        if len(self.dataset):
            del self.dataset.files[0]


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException, IPostProcessor
from GangaCore.GPIDev.Schema import Schema, SimpleItem, Version
from GangaCore.Utility.logging import getLogger
import GangaCore.Utility.Config
config = GangaCore.Utility.Config.getConfig('Configuration')
import smtplib
import email
import getpass
logger = getLogger()
# set the checkers config up


class Notifier(IPostProcessor):

    """
    Object which emails a user about jobs status are they have finished. The default behaviour is to email when a job has failed or when a master job has completed.
    Notes: 
    * Ganga must be running to send the email, so this object is only really useful if you have a ganga session running the background (e.g. screen session).
    * Will not send emails about failed subjobs if autoresubmit is on.
    """
    _schema = Schema(Version(1, 0), {
        'verbose': SimpleItem(defvalue=False, doc='Email on subjob completion'),
        'address': SimpleItem(defvalue='', doc='Email address', optional=False)
    })
    _category = 'postprocessor'
    _name = 'Notifier'
    order = 3

    def execute(self, job, newstatus):
        """
        Email user if:
        * job is a master job, or
        * job has failed but do not have auto resubmit
        * job has not failed but verbose is set to true
        """
        if len(job.subjobs) or (newstatus == 'failed' and job.do_auto_resubmit is False) or (newstatus != 'failed' and self.verbose is True):
            return self.email(job, newstatus)
        return True

    def email(self, job, newstatus):
        """
        Method to email a user about a job
        """
        sender = 'project-ganga-developers@cern.ch'
        receivers = self.address

        subject = 'Ganga Notification: Job(%s): %s has %s. (user: %s)' % (job.fqid, job.name, newstatus, getpass.getuser())
        msg_string = """
Dear User (%s),\n
Job(%s), name : %s ,  has gone into %s state.\n
Regards,
Ganga\n
PS: This is an automated notification from Ganga, 
if you would like these messages to stop please 
remove the notifier object from future jobs.
        """ % (getpass.getuser(), job.fqid,job.name, newstatus)
        msg = email.message_from_string(msg_string)
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = receivers
        string_message = msg.as_string()
        try:
            smtpObj = smtplib.SMTP(config['SMTPHost'])
            smtpObj.sendmail(sender, receivers, string_message)

        except smtplib.SMTPException as e:
            raise PostProcessException(str(e))
        return True




_global_queues = None
_queues_interface = None

def getQueues():
    return _global_queues

def startUpQueues(my_interface=None):
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    global _global_queues
    global _queues_interface
    if not my_interface:
        import GangaCore.GPI
        my_interface = GangaCore.GPI
    _queues_additional = my_interface
    if _global_queues is None:
        logger.debug("Starting Queues")
        # start queues
        from GangaCore.Runtime.GPIexport import exportToInterface
        from GangaCore.Core.GangaThread.WorkerThreads.ThreadPoolQueueMonitor import ThreadPoolQueueMonitor
        _global_queues = ThreadPoolQueueMonitor()
        exportToInterface(my_interface, 'queues', _global_queues, 'Objects')

    else:
        logger.error("Cannot Start queues if they've already started")

def shutDownQueues():
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    logger.debug("Shutting Down Queues system")
    global _global_queues
    global _queues_interface
    try:
        if _global_queues:
            _global_queues.freeze()
            _global_queues._purge_all()
            _global_queues._stop_all_threads()
    except:
        logger.warning("Error in shutting down queues thread. Likely harmless")
    _global_queues = None
    if _queues_interface:
        if hasattr(_queues_interface, 'queues'):
            delattr(_queues_interface, 'queues')
    _queues_interface = None




try:
    import pickle as pickle
except:
    import pickle

from GangaCore.Utility.logging import getLogger

logger = getLogger()

def from_file(fobj):
    return (pickle.load(fobj), [])


def to_file(obj, fileobj, ignore_subs=''):
    try:
        pickle.dump(obj, fileobj, 1)
    except Exception as err:
        logger.error("Failed to Write: %s" % obj)
        logger.error("Err: %s" % err)
        raise



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import inspect
import GangaCore.Utility.Virtualization
from GangaCore.Core.Sandbox.WNSandbox import PYTHON_DIR
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_settings, API_nullifier
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.RTHandlers.RunTimeHandlerUtils import master_sandbox_prepare, sandbox_prepare, script_generator
from GangaCore.GPIDev.Lib.File.LocalFile import LocalFile
from GangaCore.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Lib.File import File, FileBuffer
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.util import unique
from GangaCore.GPIDev.Base.Proxy import isType, stripProxy
from GangaLHCb.Lib.RTHandlers.RTHUtils import lhcbdiracAPI_script_template, lhcbdirac_outputfile_jdl
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import diracAPI_script_template, dirac_outputfile_jdl
logger = getLogger()
config = getConfig('DIRAC')
configLHCb = getConfig('LHCb')
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class ExeDiracRTHandler(IRuntimeHandler):

    """The runtime handler to run plain executables on the Dirac backend"""

    def master_prepare(self, app, appmasterconfig):
        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig)
        if type(app.exe) == File:
            input_dir = app.getJobObject().getInputWorkspace().getPath()
            exefile = os.path.join(input_dir, os.path.basename(app.exe.name))
            if not os.path.exists(exefile):
                msg = 'Executable: "%s" must exist!' % str(exefile)
                raise ApplicationConfigurationError(None, msg)

            os.system('chmod +x %s' % exefile)
        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)
        input_data,   parametricinput_data = dirac_inputdata(app)
#        outputdata,   outputdata_path      = dirac_ouputdata(app)

        job = stripProxy(app).getJobObject()
        outputfiles = [this_file for this_file in job.outputfiles if isType(this_file, DiracFile)]

        #Grab the platform if the app has that attribute
        platform = 'ANY'
        if hasattr(app, 'platform'):
            platform = app.platform

        #We have to specify a platform for the WN
        if platform == 'ANY':
            platform = configLHCb['defaultPlatform'] 
            logger.warning('No application platform specified. Using default platform %s' % platform)

        commandline = []
        commandline.append(app.exe)
        if isType(app.exe, File):
            #logger.info("app: %s" % str(app.exe.name))
            #fileName = os.path.join(get_share_path(app), os.path.basename(app.exe.name))
            #logger.info("EXE: %s" % str(fileName))
            #inputsandbox.append(File(name=fileName))
            inputsandbox.append(app.exe)
            commandline[0]=os.path.join('.', os.path.basename(app.exe.name))
        commandline.extend([str(arg) for arg in app.args])
        logger.debug('Command line: %s: ', commandline)

        #exe_script_path = os.path.join(job.getInputWorkspace().getPath(), "exe-script.py")
        exe_script_name = 'exe-script.py'

        logger.debug("Setting Command to be: '%s'" % repr(commandline))

        contents = exe_script_template()

        virtualization = job.virtualization
        if virtualization:
            contents = virtualization.modify_script(exe_script_template(), sandbox=True)

            virtualizationutils = File(inspect.getsourcefile(GangaCore.Utility.Virtualization), subdir=PYTHON_DIR )
            inputsandbox.append(virtualizationutils)

        contents = script_generator(contents,
                                    COMMAND=repr(commandline),
                                    PYTHONDIR=repr(PYTHON_DIR),
                                    OUTPUTFILESINJECTEDCODE=getWNCodeForOutputPostprocessing(job, ''))
            
        inputsandbox.append(FileBuffer(name=exe_script_name, contents=contents, executable=True))

        logger.debug("Script is: %s" % str(contents))

        from os.path import abspath, expanduser

        for this_file in job.inputfiles:
            if isinstance(this_file, LocalFile):
                for name in this_file.getFilenameList():
                    inputsandbox.append(File(abspath(expanduser(name))))

        lhcbdirac_outputfiles = lhcbdirac_outputfile_jdl(outputfiles)
        # NOTE special case for replicas: replicate string must be empty for no
        # replication
        dirac_script = script_generator(lhcbdiracAPI_script_template(),
                                        DIRAC_IMPORT='from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb',
                                        DIRAC_JOB_IMPORT='from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob',
                                        DIRAC_OBJECT='DiracLHCb()',
                                        JOB_OBJECT='LHCbJob()',
                                        NAME=mangle_job_name(app),
                                        # os.path.basename(exe_script_path),
                                        EXE=exe_script_name,
                                        # ' '.join([str(arg) for arg in app.args]),
                                        EXE_ARG_STR='',
                                        EXE_LOG_FILE='Ganga_Executable.log',
                                        ENVIRONMENT=None,  # app.env,
                                        INPUTDATA=input_data,
                                        PARAMETRIC_INPUTDATA=parametricinput_data,
                                        OUTPUT_SANDBOX=API_nullifier(outputsandbox),
                                        OUTPUTFILESSCRIPT=lhcbdirac_outputfiles,
                                        OUTPUT_PATH="",  # job.fqid,
                                        SETTINGS=diracAPI_script_settings(app),
                                        DIRAC_OPTS=job.backend.diracOpts,
                                        PLATFORM=platform,
                                        REPLICATE='True' if config['ReplicateOutputData'] else '',
                                        # leave the sandbox for altering later as needs
                                        # to be done in backend.submit to combine master.
                                        # Note only using 2 #s as auto-remove 3
                                        INPUT_SANDBOX='##INPUT_SANDBOX##'
                                        )

        #logger.info("dirac_script: %s" % dirac_script)

        #logger.info("inbox: %s" % str(unique(inputsandbox)))
        #logger.info("outbox: %s" % str(unique(outputsandbox)))

        return StandardJobConfig(dirac_script,
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def exe_script_template():
    script_template = """#!/usr/bin/env python2
'''Script to run Executable application'''
import sys, os
import subprocess
from os import system, environ, pathsep, getcwd, listdir, path

runenv = environ.copy()
if 'PATH' in runenv.keys():
    runenv['PATH'] = getcwd() + (pathsep + runenv['PATH'])
else:
    runenv['PATH'] = getcwd()
if not 'PWD' in runenv.keys():
    runenv['PWD'] = getcwd()

PYTHON_DIR = ###PYTHONDIR###
workdir = getcwd()

sys.path.insert(0,path.join(workdir, PYTHON_DIR))

execmd = ###COMMAND###

if isinstance(execmd, str):
    execmd = [execmd]

if isinstance(execmd, list):
    if path.isfile(path.abspath(execmd[0])):
        execmd[0] = path.abspath(execmd[0])

###VIRTUALIZATION###

err = None
try:
    rc = subprocess.call(execmd, env=runenv, shell=False)
except Exception as x:
    rc = -9999
    print('Exception occured in running process: ' + repr(execmd))
    print('Err was: ' + str(x))
    subprocess.call('echo $PATH', shell=True)
    print('PATH: ' + str(runenv['PATH']))
    print('PWD: ' + str(runenv['PWD']))
    print("files on WN: " + str(listdir('.')))
    raise

print("files on WN: " + str(listdir('.')))

###OUTPUTFILESINJECTEDCODE###

try:
    subprocess.call('chmod -R u+rwX  * .*', env=runenv, shell=True)
except Exception as x:
    print('Problems fixing permissions on output files to allow for clean-up.')
    print('Err was: ' + str(x))

sys.exit(rc)
"""
    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('Executable', 'Dirac', ExeDiracRTHandler)



################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PrimeFactorizer.py,v 1.2 2009-04-23 13:47:59 moscicki Exp $
################################################################################

import os
import shutil
from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Lib.File import File, ShareDir
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import getName

logger = getLogger()

class PrimeFactorizer(IPrepareApp):
    """
    PrimeFactorizer application -- factorize any arbitrary number into prime numbers.
    """

    _schema = Schema(Version(1, 1), {
        'number': SimpleItem(defvalue=1, sequence=0, typelist=['int', 'long'], doc='The number to be factorized.'),
        'is_prepared': SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, hidden=0, typelist=[None, ShareDir], protected=0, comparable=1, doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        'hash': SimpleItem(defvalue=None, typelist=[None, str], hidden=0, doc='MD5 hash of the string representation of applications preparable attributes'),
    })

    _category = 'applications'
    _name = 'PrimeFactorizer'
    _exportmethods = ['prepare', 'unprepare']
    # _GUIPrefs = [ { 'attribute' : 'number', 'widget' : 'Integer' },
    #              { 'attribute' : 'prime_table', 'widget' : 'String' }]

    def __init__(self):
        super(PrimeFactorizer, self).__init__()

    def configure(self, masterappconfig):
        from GangaCore.GPI import TUTDIR
        from GangaCore.Core.exceptions import ApplicationConfigurationError
        import os.path

        # the prime number factorizer executable
        self.exe = File(TUTDIR + '/Lib/primes/prime_factor.py')

        prime_tables = []
        try:
            job = self.getJobObject()
            prime_tables = job.inputdata.get_dataset()
        except:
            pass

        # the arguments of the executable
        # - arg1: the number
        # - arg2: the SURL of the prime number lookup table
        self.args = ['%s' % self.number] + prime_tables
        self.inputs = []
        self.outputs = ['*.dat']
        self.envs = {}
        return (None, None)

    def unprepare(self, force=False):
        """
        Revert an Executable() application back to it's unprepared state.
        """
        logger.debug('Running unprepare in Executable app')
        if self.is_prepared is not None:
            self.decrementShareCounter(self.is_prepared)
            self.is_prepared = None
        self.hash = None

    def prepare(self, force=False):
        from GangaCore.Core.exceptions import ApplicationPrepareError

        if (self.is_prepared is not None) and not force:
            raise ApplicationPrepareError(
                '%s application has already been prepared. Use prepare(force=True) to prepare again.' % getName(self))

        # lets use the same criteria as the configure() method for checking file existence & sanity
        # this will bail us out of prepare if there's somthing odd with the job config - like the executable
        # file is unspecified, has a space or is a relative path
        self.configure(self)
        logger.info('Preparing %s application.' % getName(self))
        self.is_prepared = ShareDir()
        logger.info('Created shared directory: %s' % (self.is_prepared.name))

        try:
            # copy any 'preparable' objects into the shared directory
            send_to_sharedir = self.copyPreparables()
            # add the newly created shared directory into the metadata system
            # if the app is associated with a persisted object
            self.checkPreparedHasParent(self)
            # return
            # [os.path.join(self.is_prepared.name,os.path.basename(send_to_sharedir))]
            self.post_prepare()

            if isinstance(self.exe, str):
                source = self.exe
            elif isinstance(self.exe, File):
                source = self.exe.name

            if isinstance(self.exe, str):
                logger.debug("exe is a string so no copying")
            elif not os.path.exists(source):
                logger.debug("Error copying exe: %s to input workspace" %
                            str(source))
            else:
                try:
                    parent_job = self.getJobObject()
                except:
                    parent_job = None
                    pass
                if parent_job is not None:
                    input_dir = parent_job.getInputWorkspace(create=True).getPath()
                    shutil.copy2(source, input_dir)

        except Exception as err:
            logger.debug("Err: %s" % str(err))
            self.unprepare()
            raise

        return 1


# FIXME: a cleaner solution, which is integrated with type information in schemas should be used automatically
#config = getConfig('GangaTutorial_Properties')
mc = getConfig('MonitoringServices')
mc.addOption('GangaTutorial', None, '')


class RTHandler(IRuntimeHandler):
    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        c = StandardJobConfig(app.exe, app.inputs, app.args, app.outputs, app.envs)
       
        return c


class LCGRTHandler(IRuntimeHandler):
    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        c = LCGJobConfig(app.exe, app.inputs, app.args, app.outputs, app.envs)
        c.monitoring_svc = mc['GangaTutorial']
        
        return c

allHandlers.add('PrimeFactorizer', 'LSF', RTHandler)
allHandlers.add('PrimeFactorizer', 'Local', RTHandler)
allHandlers.add('PrimeFactorizer', 'PBS', RTHandler)
allHandlers.add('PrimeFactorizer', 'SGE', RTHandler)
allHandlers.add('PrimeFactorizer', 'Condor', RTHandler)
allHandlers.add('PrimeFactorizer', 'LCG', LCGRTHandler)
allHandlers.add('PrimeFactorizer', 'TestSubmitter', RTHandler)
allHandlers.add('PrimeFactorizer', 'Interactive', RTHandler)
allHandlers.add('PrimeFactorizer', 'Batch', RTHandler)

##############################################################
# handler for NG

class NGRTHandler(IRuntimeHandler):

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        from GangaNG.Lib.NG import NGJobConfig
        return NGJobConfig(app.exe, app.inputs, app.args, app.outputs, app.envs)
